/*									tab:8
 *
 * nxam.c - Functions that make NX look like active messages.
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Steve Luna
 * Version:		71
 * Creation Date:	Wed Oct 13 12:46:17 1993
 * Filename:		nxam.c
 * History:
 *	LTL	67	Sun Jul 23 12:48:01 1995
 *		Changed am_request_fd() and am_reply_fd() to am_request_df() and
 *    am_reply_df(), respectively, to make them consistent with GAM spec 1.1.
 *    Added calls to gsync() in am_enable() and am_disable().
 *	LTL	20	Sun Feb 19 10:35:29 1995
 *		Made it GAM v1.1 compliant
 *	ssl	11	Tue Apr 19 12:28:39 1994
 *		fixed bug initializing NXAM_free_seg
 */


#include <stdio.h>
#include <limits.h>
#include <nx.h>

typedef unsigned int vnn_t;
typedef	void (*i_handler_t)(vnn_t src, int arg1, int arg2, int arg3, int arg4);
typedef	void (*d_handler_t)(vnn_t src, int arg1, int arg2, double arg3);
typedef void (*xfer_handler_t)(vnn_t src_node, void *buf, int len,
															 void *info);
		
#define NUM_seg 256
#define NUM_irecv 1
#define AM_INT 0
#define AM_DOUBLE 1
#define NXAM_msg_type NUM_seg

/* use these to get unique message types. since messages to a given processor
   are ordered, only need separate ones for sending (put, store) and recving
   (get) - ssl */

#define SENDTYPE(proc) (proc)
#define RECVTYPE(proc) (am_numproc+(proc))
				
typedef struct {
  i_handler_t handler;
	int type;
	vnn_t src_node; 
	int args[4];
} nxam_pkt;


typedef struct {
	xfer_handler_t end_fun;
	void *buf;
	int len;
	void *info;
	int mid;
	int src_node;
} xfer_segment;

static nxam_pkt nxam_buf[NUM_irecv];
static int mid[NUM_irecv];
static int am_myproc;
static int am_numproc;
static xfer_segment NXAM_seg[NUM_seg];
static int NXAM_free_seg[NUM_seg/32];
static char fho[256];	/* table to find #shifts to highest order bit of byte */

void am_enable()
{
	int i, b;
	for (i = 0; i < NUM_irecv; i++)
		mid[i] = irecv(NXAM_msg_type, &nxam_buf[i], sizeof(nxam_pkt));

	for (i = 0; i < NUM_seg/32; i++)
		NXAM_free_seg[i] = -1;
	fho[0] = -1;
	for (i = 1; i<256; i++)
	{
		b = 7;
		while(! (i&(1<<b))) b--;
		fho[i] = b;
	}
	am_myproc = mynode();
	am_numproc = numnodes();
	am_barrier_init();
	gsync();
}

void am_disable()
{
  gsync();
}


int am_my_proc()
{
  return am_myproc;
}

int am_procs()
{
  return am_numproc;
}

int am_max_size()
{
  return INT_MAX;
}

void am_poll()
{
	union {
   double d;
	 int i[2]; 
	} u;		
	int i, j;
	xfer_segment seg;
	static int in_handler = 0;
	int *x;
	int received;

	if (in_handler)
	{
		fprintf(stderr, "FATAL error: "
				"am_poll called from within a handler on proc %d\n",
				mynode());
		/*exit(1);*/
		/* cause core dump to find error with ipd */
		x = 0;
		*x = 5;
	}
	received = 1;

  /* keep draining until nothing left; otherwise we may deadlock
	 * this increases the overhead a bit becuase we will poll at least once
	 * more if we get something -- ltl
	 */
	while (received) {
		received = 0;	
		/* poll for small messages */
		for (i = 0; i < NUM_irecv; i++)
		{

			if (msgdone(mid[i])) {
				in_handler = 1;
				if (nxam_buf[i].type == AM_INT) {
					((i_handler_t)nxam_buf[i].handler)(nxam_buf[i].src_node,
														nxam_buf[i].args[0], nxam_buf[i].args[1],
														nxam_buf[i].args[2], nxam_buf[i].args[3]);
				}
				else {
					u.i[0] = nxam_buf[i].args[0];
					u.i[1] = nxam_buf[i].args[1];        					
					((d_handler_t)nxam_buf[i].handler)(nxam_buf[i].src_node,  
														nxam_buf[i].args[2], nxam_buf[i].args[3], u.d);

				}
				received = 1;
				in_handler = 0;
				mid[i] = irecv(NXAM_msg_type, &nxam_buf[i], sizeof(nxam_pkt));
			}
		}

		/* poll for bulk messages */
		for (i = (NUM_seg/32)-1; i>=0; i--)
			if (NXAM_free_seg[i] != -1)
				for (j = 31; j>=0; j--)
					if ((NXAM_free_seg[i]&(1<<j)) == 0)
						if (msgdone((seg=NXAM_seg[(i<<5)+j]).mid)) {
							NXAM_free_seg[i] ^= (1<<j);
							((xfer_handler_t)seg.end_fun)(seg.src_node, seg.buf,
																						seg.len, seg.info);
							received = 1;
						}
  }
}

void am_wait(volatile int *flag, int value)
{
	while (*flag < value)
	{
		am_poll();
	}
	(*flag)-=value;
}

void am_poll_wait(volatile int *flag, int value)
{
	do
		am_poll();
	while (*flag < value);
	(*flag)--;
}

void am_request_4(int dest, void *fun, int arg1, int arg2, int arg3, int arg4)
{
	nxam_pkt nxam_sbuf;

	nxam_sbuf.handler = fun;
	nxam_sbuf.src_node = am_myproc;
	nxam_sbuf.type = AM_INT;
	nxam_sbuf.args[0] = arg1;
	nxam_sbuf.args[1] = arg2;
	nxam_sbuf.args[2] = arg3;
	nxam_sbuf.args[3] = arg4;
	csend(NXAM_msg_type, &nxam_sbuf, sizeof(nxam_pkt), dest, myptype());
	am_poll();
}


void am_request_df(int dest, void *fun, int arg1, int arg2, double arg3)
{
	nxam_pkt nxam_sbuf;
	union {
   double d;
	 int i[2]; 
	} u;
	
	nxam_sbuf.handler = fun;
	nxam_sbuf.src_node = am_myproc;
	nxam_sbuf.type = AM_DOUBLE;
	u.d = arg3;
	nxam_sbuf.args[0] = u.i[0];
	nxam_sbuf.args[1] = u.i[1];
	nxam_sbuf.args[2] = arg1;
	nxam_sbuf.args[3] = arg2;
	csend(NXAM_msg_type, &nxam_sbuf, sizeof(nxam_pkt), dest, myptype());
	am_poll();
}

void am_reply_4(int dest, void *fun, int arg1, int arg2, int arg3, int arg4)
{
	nxam_pkt nxam_sbuf;

	nxam_sbuf.handler = fun;
	nxam_sbuf.src_node = am_myproc;
	nxam_sbuf.type = AM_INT;	
	nxam_sbuf.args[0] = arg1;
	nxam_sbuf.args[1] = arg2;
	nxam_sbuf.args[2] = arg3;
	nxam_sbuf.args[3] = arg4;
	csend(NXAM_msg_type, &nxam_sbuf, sizeof(nxam_pkt), dest, myptype());
}


void am_reply_df(int dest, void *fun, int arg1, int arg2, double arg3)
{
	nxam_pkt nxam_sbuf;
	union {
   double d;
	 int i[2]; 
	} u;
	
	nxam_sbuf.handler = fun;
	nxam_sbuf.src_node = am_myproc;
	nxam_sbuf.type = AM_DOUBLE;
	u.d = arg3;
	nxam_sbuf.args[0] = u.i[0];
	nxam_sbuf.args[1] = u.i[1];
	nxam_sbuf.args[2] = arg1;
	nxam_sbuf.args[3] = arg2;
	csend(NXAM_msg_type, &nxam_sbuf, sizeof(nxam_pkt), dest, myptype());
}

int allocate_segment()
{
	int i, b, f, free;
	for (i = (NUM_seg/32)-1; i>=0; i--)
	{
		if (NXAM_free_seg[i])
		{
			free = NXAM_free_seg[i];
			if (f=(free>>24)&0xff) {
				b = 24+fho[f];
			} else if (f=(free>>16)&0xff) {
				b = 16+fho[f];
			} else if (f=(free>>8)&0xff) {
				b = 8+fho[f];
			} else {
				b = fho[free&0xff];
			}
		  NXAM_free_seg[i] = free ^ (1<<b);
			return (i<<5)+b;
		}
	}
	fprintf(stderr, "error in nxam.c, couldn't allocate segment\n");
}


void am_open_send_segment(int src_node, void *base, int count, void *fun,
										 void *info)
{
	int seg;
	
	seg = allocate_segment();
	NXAM_seg[seg].mid = irecv(SENDTYPE(src_node), base, count);
	NXAM_seg[seg].end_fun = fun;
	NXAM_seg[seg].info = info;
	NXAM_seg[seg].len = count;
	NXAM_seg[seg].buf = base;
	NXAM_seg[seg].src_node = src_node;	
}

void am_open_receive_segment(int src_node, void *base, int count, void *fun,
										 void *info)
{
	int seg;
	
	seg = allocate_segment();
	NXAM_seg[seg].mid = irecv(RECVTYPE(src_node), base, count);
	NXAM_seg[seg].end_fun = fun;
	NXAM_seg[seg].info = info;
	NXAM_seg[seg].len = count;
	NXAM_seg[seg].buf = base;
	NXAM_seg[seg].src_node = src_node;	
}

void am_xfer(int dest, int seg, void *buf, int size)
{
	csend(seg, buf, size, dest, myptype());
	am_poll();
}

void am_reply_xfer(int dest, int seg, void *buf, int size)
{
	csend(seg, buf, size, dest, myptype());
}

/* This used to be  __store() in Split-C -- ltl */
int am_store(vnn_t dest, void *src_buf, void *dest_buf, int len,
							xfer_handler_t request_handler, void *handler_arg)
{
	if (dest == am_myproc) {
		memmove(dest_buf, src_buf, len);
		/* invoke the handler */
		(request_handler)(dest, dest_buf, len, handler_arg);
	}
	else {
		am_request_4(dest, am_open_send_segment, (int)dest_buf, len,
								 (int)request_handler, (int)handler_arg);
		am_xfer(dest, SENDTYPE(am_myproc), src_buf, len);
	}
}

/* Could have used non-blocking send to implement this. Not sure it
 * is better though -- ltl
 */
int am_store_async(vnn_t dest, void *src_buf, void *dest_buf,
									 int len, xfer_handler_t request_handler,
									 void *request_handler_arg,
									 xfer_handler_t endfunc, void *endfunc_arg)
{
  am_store(dest, src_buf, dest_buf, len, request_handler,
					 request_handler_arg);
	/* invoke the handler */
	(endfunc)(dest, src_buf, len, endfunc_arg);
}


/* This used to be __get_ctr() in Split-C -- ltl */
int am_get(vnn_t dest, void *src_buf, void *dest_buf, int len,
					 xfer_handler_t reply_handler, void *handler_arg)
{
	int segId;

	if (dest == am_myproc) {
    memmove(dest_buf, src_buf, len);
		/* invoke the handler */
		(reply_handler)(dest, dest_buf, len, handler_arg);		
	} 
	else {
		/* Set up a segment */
		am_open_receive_segment(dest, dest_buf, len,
								 reply_handler, handler_arg);			
		/* Call the function in the remote node which does the xfer */
 		am_request_4(dest, (void *)am_reply_xfer, RECVTYPE(dest), (int)src_buf,
					len, 0);
	}
}

