/*                                                                      tab:2
 * tcpam.c - Active Messages on top of TCP/IP
 *
 * "Copyright (c) 1994 by Lok Tin Liu and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:                      Lok Tin Liu
 * Version:                     456
 * Creation Date:       Fri Jul  8 13:55:26 1994
 * Filename:            tcpam.c
 * History:
 *      LTL     1       Fri Nov 11 15:49:16 1994
 *              Incorporated fixes by Chad and Alan
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <values.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#ifndef __hpux
#include <sys/select.h>
#endif
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include "tcpam_int.h"
#include <fcntl.h>
#include <errno.h>

/* GLUnix interface */
#include <cam.h>
#include <glib.h>

#define BASE_PORT 11000	/* base of server port */
#define MAX_HOSTNAME_LEN 256
char my_hostname[MAX_HOSTNAME_LEN];
#define MAX_RETRY 60

#define QUICK 1

#ifndef OPEN_MAX
#define OPEN_MAX 16
#endif

#define _BSD 44			/* for AIX */

/* maximum length of hostname */
#define MAX_HOSTNAME_LEN 256


/* message tags */
#define TCPAM_AM_INT_PKT 0
#define TCPAM_AM_DOUBLE_PKT 1
#define TCPAM_STORE_PKT 2
#define TCPAM_GET_REQUEST_PKT 3

#define WAIT_HEADER 0
#define WAIT_DATA   1

/*Number of connections between peers */
/*One for send and one for receive... */
#define TCPAM_NUM_CONNECTIONS 2
#define REQ_CONN 0
#define REPLY_CONN 1
/* Hardcoded for now, but these N connections are the building blocks
 * for MT-Safe GAM... COY
 */
#ifndef MAX
#define MAX(a,b) (a>b)?a:b;
#endif

typedef void (*TCPAM_handler_4i) (vnn_t calling_node,
				  int arg1, int arg2, int arg3, int arg4);
typedef void (*TCPAM_handler_2i1d) (vnn_t calling_node,
				    double arg1, int arg2, int arg3);
typedef void (*TCPAM_xfer_endfunc) (vnn_t src_node, void *addr, int nbytes,
				    void *handler_arg);

typedef struct {
    int tag;
    vnn_t calling_node;
    handler_t handler;
    int arg1, arg2, arg3;
    union {
	int arg4, arg5;
	double d_arg4;
    } opt_arg;
} TCPAM_pkt;

typedef struct {
    int tag;
    handler_t endfunc;
    int arg1, arg2, arg3, arg4;
    void *addr;
    int len;
} TCPAM_xfer_header;

int base_port;
int AM_myproc;
int AM_procs;
int AM_errno;

/* server port to initiate connections */
int TCPAM_server_port[TCPAM_NUM_CONNECTIONS];

/* used with select() */
struct timeval TCPAM_timeval;

/* maximum file descriptor number per connection */
int TCPAM_max_sockfd[TCPAM_NUM_CONNECTIONS];

/* bit vector used with select() */
static fd_set TCPAM_fd_set[TCPAM_NUM_CONNECTIONS], TCPAM_all_fd_set;

/* can have as many processors as the max. number of file open */
struct sockaddr_in host_entries[OPEN_MAX];

/* corresponding socket descriptor to each node */
int node_sockfd[OPEN_MAX][TCPAM_NUM_CONNECTIONS];

/* message headers for request and reply sockets to each node */
TCPAM_pkt pkt_headers[OPEN_MAX][TCPAM_NUM_CONNECTIONS];

/* message state for request and reply sockets to each node */
int message_state[OPEN_MAX][TCPAM_NUM_CONNECTIONS];

/* pending data bytes for request and reply sockets to each node */
int msg_bytes_left[OPEN_MAX][TCPAM_NUM_CONNECTIONS];

/* pending data bytes for message headers ... */
int hdr_bytes_read[OPEN_MAX][TCPAM_NUM_CONNECTIONS];

/* hostname of each node */
char node_hostname[OPEN_MAX][MAX_HOSTNAME_LEN];

int retry = 0;

/* get the hostnames and IP addresses in the partition from GLUNIX */
void TCPAM_read_hostnames(int num_procs)
{
    int i;
    char hostname[MAX_HOSTNAME_LEN];
    struct in_addr inaddr, *ip;
    struct hostent *hp;

    if (num_procs > OPEN_MAX) {
	printf("Cannot start program with %d hosts (%d max)\n", num_procs,
	       OPEN_MAX);
	exit(-1);
    }
    for (i = 0; i < num_procs; ++i) {
	ip = (struct in_addr *) Am_GetFastAddrByVnn(i);
	if (ip) {
	    host_entries[i].sin_family = AF_INET;
	    host_entries[i].sin_port = 0;

	    host_entries[i].sin_addr.s_addr = ip->s_addr;
#ifdef DEBUG 
	    printf("TCPAM: node %d read node %d IP address %x\n",
						 AM_myproc, i, ip->s_addr);
      fflush(stdout);
#endif			
	    hp = gethostbyaddr((char *) ip, sizeof(struct in_addr), AF_INET);
	    if ((hp == 0) || (hp->h_name == 0)) {
     		printf("Cannot get hostname for IP address %x %x\n",	*ip, ip);
		exit(-1);
	    }
	    strcpy(node_hostname[i], hp->h_name);
	} else {
	    printf("Cannot get IP address for node %d from GLUNIX\n", i);
	    exit(-1);
	}
    }
 /* stagger the nodes so that they won't swarm in to make connections */		
 /*	sleep((unsigend)(AM_myproc*0.1)); */
}

/* get configuration from GLUNIX */
void TCPAM_get_config()
{
    pid_t pid;
    Npid glunix_pid;
		int ns;
    char *base_port_ptr;
		
    /* get total number of nodes */
    AM_procs = Glib_GetParallelDegree();

    if (AM_procs <= 0) {
	printf("GLUNIX reports that this proces is not part of a parallel program: AM_procs = %d\n", AM_procs);
	exit(-1);
    }
    /* get my node number */
    AM_myproc = Glib_GetMyVnn();
#ifdef DEBUG
		printf("TCPAM: %s is node %d\n", my_hostname, AM_myproc);
#endif		
    pid = getpid();
    if (pid < 0) {
	perror("Cannot get local PID");
	exit(-1);
    }
    /* get the GLUNIX pid to agree on a server port number */
    glunix_pid = Glib_GetMyNpid();
#ifdef DEBUG 
		printf("TCPAM: node %d is running NPID %d\n", AM_myproc, glunix_pid);
#endif				
    if (glunix_pid == NPID_NOBODY) {
	printf("GLUNIX reports that this proces is not part of a parallel program: glunix_pid = %d\n", glunix_pid);
	exit(-1);
    }


		/* if TCPAM_BASE_PORT environment variable not defined */
		if ((base_port_ptr = getenv("TCPAM_BASE_PORT")) == (char *) 0) {
      base_port = BASE_PORT;
		}
		else {
     base_port = atoi(base_port_ptr);
		}
/*		printf("TCPAM: node %d is using port %d\n", AM_myproc, base_port); */
  	for (ns = 0; ns < TCPAM_NUM_CONNECTIONS; ns++)	
			TCPAM_server_port[ns] = base_port + glunix_pid + 1000*ns;
    TCPAM_read_hostnames(AM_procs);
}


/* set various socket options */
void TCPAM_set_socket(int sockfd)
{
    int opt = 1;
    int opt_len = sizeof(opt);

    /* don't delay send to buffer packets */
    if (setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY, (void *)&opt,
		   opt_len) < 0) {
	perror("setsockopt TCP_NODELAY failed");
	close(sockfd);
	exit(-1);
    }
    /* allow local address to be reused */
    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (void *)&opt,
		   opt_len) < 0) {
	perror("setsockopt SO_REUSEADDR failed");
	close(sockfd);
	exit(-1);
    }
    opt = SB_MAX;
    /* set send and receive buffer size to maximum */
    if (setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, (void *)&opt,
		   opt_len) < 0) {
	perror("setsockopt SO_RCVBUF failed");
	close(sockfd);
	exit(-1);
    }
    if (setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, (void *)&opt,
		   opt_len) < 0) {
	perror("setsockopt SO_SNDBUF failed");
	close(sockfd);
	exit(-1);

    }
}


/* search for the corresponding node using the IP address and store the
 * socket file descriptor to the connection_num
 */
void TCPAM_store_sockfd(struct sockaddr_in *client_addr, int client_sockfd,
			int connection_num)
{
    int i, found;

    found = 0;
    /* use linear search -- hope it won't be too slow */
    i = 0;
    while ((i < AM_procs) && (!found)) {
	if (host_entries[i].sin_addr.s_addr == client_addr->sin_addr.s_addr)
	    found = 1;
	else
	    ++i;
    }
    if (found)
	node_sockfd[i][connection_num] = client_sockfd;
    else {
	printf("Error: cannot find IP address %s in host table\n",
	       inet_ntoa(client_addr->sin_addr));
	am_exit(-1);
    }
#ifdef DEBUG	
    printf("TCPAM: node %d match socket %d for node %d (connection %d)\n",
					 AM_myproc, client_sockfd, i, connection_num);
#endif		
}


/* open known server port socket for receiving connection requests 
 */
int TCPAM_open_server(int conn)
{
    int my_sockfd, flags;
    struct sockaddr_in my_addr;

    /* first open TCPAM_server_port as server port to set up connections */
    if ((my_sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	perror("Cannot open server stream socket");
	exit(-1);
    }
    my_addr.sin_family = AF_INET;
    my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    my_addr.sin_port = htons(TCPAM_server_port[conn]);

    TCPAM_set_socket(my_sockfd);
#ifdef DEBUG
    printf("my_sockfd = %d\n", my_sockfd);
    printf("s_addr = %d\n", my_addr.sin_addr.s_addr);
    printf("s_port = %d\n", my_addr.sin_port);
    printf("TCPAM_server_port = %d\n", TCPAM_server_port[conn]);
#endif
    if (bind(my_sockfd, (struct sockaddr *) &my_addr,
	     sizeof(my_addr)) < 0) {

	perror("Cannot bind local address to server port");
	perror("");
	close(my_sockfd);
	exit(-1);
    }
		/* set socket to non-blocking */
#if 0		
 		flags = fcntl(my_sockfd, F_GETFL);
 		flags |= O_NONBLOCK;
 		fcntl(my_sockfd, F_SETFL, flags);
#endif
#ifdef DEBUG
		printf("node %d open server port %d fd %d\n", AM_myproc,
					 TCPAM_server_port[conn], my_sockfd);
#endif		
    return (my_sockfd);
}


/* open sockets and connect to all other nodes on connection number 
   connection_num.  Use my_sockfd to make the connection.
 */
void TCPAM_open_sockets(int num_procs, int connection_num, int my_sockfd)
{
    int i, addr_len, connected, retry;
    struct sockaddr_in server_addr, client_addr;
    int client_sockfd;
#ifdef DEBUG
    printf("TCPAM: node %d TCPAM_open_sockets num %d conn %d fd %d\n",
					 AM_myproc, num_procs,
	   connection_num, my_sockfd);
    fflush(stdout);
#endif		
    /* initialize socket descriptor table */
    for (i = 0; i < OPEN_MAX; ++i)
	node_sockfd[i][connection_num] = -1;

    TCPAM_max_sockfd[connection_num] = -1;
    addr_len = sizeof(struct sockaddr);


    node_sockfd[AM_myproc][connection_num] = my_sockfd;		/*for loopback? */

    listen(my_sockfd, 5);
/*    printf("TCPAM_open_sockets 2\n");
    fflush(stdout); */
    /* make connections to all nodes with node number < AM_myproc */
    for (i = 0; i < AM_myproc; ++i) {
    	server_addr.sin_addr.s_addr =
			host_entries[i].sin_addr.s_addr;
    	server_addr.sin_port = htons(TCPAM_server_port[connection_num]);
	
#ifdef DEBUG
	printf("TCPAM: node %d trying to connect to node %d\n", AM_myproc, i);
#endif	

	if ((client_sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	    printf("Cannot open client stream socket\n");
	    am_exit(-1);
	}
	TCPAM_set_socket(client_sockfd);
	retry = 0;
	connected = 0;
	/* Keep retrying if server is not ready; can we livelock here? */
	while ((!connected) && (retry < MAX_RETRY)) {
#ifdef DEBUG		
	    printf("TCPAM: node %d TCPAM_open_sockets retrying %d times node %d port %d\n",
						 AM_myproc, retry, 	i, server_addr.sin_port);
	    fflush(stdout);
#endif			
	    if (connect(client_sockfd, (struct sockaddr *) &server_addr,
			sizeof(server_addr)) < 0) {
		close(client_sockfd);
		sleep(1);
		server_addr.sin_family = AF_INET;
		server_addr.sin_addr.s_addr = host_entries[i].sin_addr.s_addr;
		server_addr.sin_port = htons(TCPAM_server_port[connection_num]);
		if ((client_sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		    printf("Cannot open client stream socket\n");
		    am_exit(-1);
		}
		++retry;
	    } else
		connected = 1;

	}

	if (!connected) {
	    printf("Cannot connect to %s after %d seconds timeout",
		   node_hostname[i], MAX_RETRY);
	    perror("");
	    close(client_sockfd);
	    am_exit(-1);
	}
	node_sockfd[i][connection_num] = client_sockfd;
#ifdef DEBUG	
	printf("TCPAM: node %d open socket %d for node %d, connection_num %d\n",
				 AM_myproc, client_sockfd, i, connection_num);
	fflush(stdout);
#endif	
	if (client_sockfd > TCPAM_max_sockfd[connection_num])
	    TCPAM_max_sockfd[connection_num] = client_sockfd;
    }

		
    /* make connections to all nodes with node number > AM_myproc */
    for (i = 1; i <= (AM_procs - AM_myproc - 1); ++i) {
			retry = 0;
#ifdef DEBUG
    printf("TCPAM: node %d TCPAM_open_sockets accepting iteration %d fd %d\n",
					 AM_myproc, i, my_sockfd);
    fflush(stdout);
#endif

			client_sockfd = accept(my_sockfd, (struct sockaddr *) &client_addr,
			       &addr_len);

#ifdef DEBUG
		if (client_sockfd == -1) {  	
    printf("TCPAM: node %d TCPAM_open_sockets iteration %d retry %d errno %d\n",
					 AM_myproc, i, retry, errno);
		perror("");
    fflush(stdout);
    }
#endif

	if ((client_sockfd == -1) || (retry >= MAX_RETRY)) {
	    printf("Error: cannot accept connection on %s\n",
		   node_hostname[AM_myproc]);
	    perror("accept");
	    am_exit(-1);
	}
	TCPAM_set_socket(client_sockfd);

	/* store the socket file descriptor of the corresponding node */
	TCPAM_store_sockfd(&client_addr, client_sockfd, connection_num);
	if (client_sockfd > TCPAM_max_sockfd[connection_num])
	    TCPAM_max_sockfd[connection_num] = client_sockfd;
    }
    ++TCPAM_max_sockfd[connection_num];		/*Make the max strictly greater than */
    /*Why do this? COY */
#ifdef DEBUG
    printf("TCPAM: node %d TCPAM_open_sockets done\n", AM_myproc);
    fflush(stdout);
#endif		
}


/* set up the TCPAM layer

 * 1. read in the configuration file
 * 2. open sockets and connect to all other nodes
 * 3. initialize data structures
 */
void am_enable()
{
    int i, b, ns, flags;
    int num_procs;
    int my_sockfd[TCPAM_NUM_CONNECTIONS];
    volatile int j;
		
#ifdef DEBUG
    if (gethostname(my_hostname, MAX_HOSTNAME_LEN) != 0) {
	     perror("Cannot get hostname");
	     exit (-1);
		}
    printf("TCPAM: %s is in am_enable()\n", my_hostname);
		fflush(stdout);
#endif
		
    TCPAM_get_config();		/* get network configuration from GLUnix */
    am_barrier_init();		/* initialize data structures for barrier */

#ifdef DEBUG
    printf("TCPAM: node %d server port %d %d\n", AM_myproc,
					 TCPAM_server_port[0], TCPAM_server_port[1]);
    fflush(stdout);		
#endif
		
      /*open the server port for receiving connection requests */
    for (ns = 0; ns < TCPAM_NUM_CONNECTIONS; ns++) {		
			my_sockfd[ns] = TCPAM_open_server(ns);				
  	}
    /* open sockets and connect to all other processors; this also acts as
     * an implicit barrier for the setup phase
     */
    for (ns = 0; ns < TCPAM_NUM_CONNECTIONS; ns++) {

    	TCPAM_open_sockets(AM_procs, ns, my_sockfd[ns]);
    	/* set up descriptor bit vector for select() */
    	FD_ZERO(&TCPAM_fd_set[ns]);
    	for (i = 0; i < AM_procs; ++i) {
	      FD_SET(node_sockfd[i][ns], &TCPAM_fd_set[ns]);
#ifdef QUICK
  	    FD_SET(node_sockfd[i][ns], &TCPAM_all_fd_set);
#endif
     	}
#ifdef DEBUG 
    printf("TCPAM: node %d is in am_enable() 3\n", AM_myproc);
    fflush(stdout);		
#endif			
    }


    /* use non-blocking select() */
    TCPAM_timeval.tv_sec = 0;
    TCPAM_timeval.tv_usec = 0;

    for (ns = 0; ns < TCPAM_NUM_CONNECTIONS; ns++) {
    	for (i = 0; i < AM_procs; ++i) {
	      if (node_sockfd[i][ns] != -1) {
      		flags = fcntl(node_sockfd[i][ns], F_GETFL);
      		flags |= O_NONBLOCK;
      		fcntl(node_sockfd[i][ns], F_SETFL, flags);
  	    }
    	}
    }

#ifdef DEBUG 
    printf("TCPAM: node %d am_enable() done\n", AM_myproc);
    fflush(stdout);		
#endif		
}

/*Or could have everyone fill in the same file descriptor set,
   and then selectively mask outs this */


/* close all sockets */
void am_disable()
{
    int i, ns;

    /*spin for a while before exit; otherwise some nodes will hang on the
     *last barrier
     */
    for (i = 0; i < 100000; ++i);

    /* close all sockets */
    for (ns = 0; ns < TCPAM_NUM_CONNECTIONS; ns++)
	for (i = 0; i < AM_procs; ++i) {
	    if (node_sockfd[i][ns] != -1)
		close(node_sockfd[i][ns]);
	}
}

/* close all sockets before exit */
void am_exit(int code)
{
    am_disable();
    exit(code);
}

/*Optimization--just go up to max */
/*OR two sets together */
#define OR_FD_SET(a,b)\
{\
   int i;\
   for(i=0; i< howmany(FD_SETSIZE, NFDBITS); i++)\
     a.fds_bits[i] |= b.fds_bits[i];\
}


/*
 * Caller blocks until 1 or more messages arrives.  Messages
 * are not actually processed.  This function is used for event
 * driven applications.
 */
int am_sleep()
{
    fd_set TCPAM_temp_fd_set;
    int sockfd, max, i;
    int num_pending;

#ifdef QUICK
    /*This  value is precomputed for quicker response times--yeah, right*/
    TCPAM_temp_fd_set = TCPAM_all_fd_set;
#else
    TCPAM_temp_fd_set= TCPAM_fd_set[REQ_CONN];
    OR_FD_SET(TCPAM_temp_fd_set, TCPAM_fd_set[REPLY_CONN]);
#endif

    max = MAX(TCPAM_max_sockfd[REQ_CONN],TCPAM_max_sockfd[REPLY_CONN]);

    /* 
     * check whether there are messages pending (note that
     * select() will put result in TCPAM_temp_fd_set) -- uses
     * blocking select() call, to block caller until msg arrives
     */ 	
    /*could we get stuck if a socket connection breaks ? */

    do {
	num_pending = select(max /*COY*/,
			     &TCPAM_temp_fd_set,
			     (fd_set *)0,
			     (fd_set *)0, 0);

	if(num_pending == -1) {
	    AM_errno = errno;
	    am_exit(-1);
	}

    } while(num_pending == 0);
   
    return(0);
}

/*
 * Poll request and reply networks. Returns -1 on encountering 
 * any error and may set AM_errno, and returns 0 otherwise.
 */
int am_poll()
{
    char *cp;
    TCPAM_pkt *pp;
    TCPAM_xfer_header *hp;
    fd_set TCPAM_temp_fd_set;

    int i, n, m, sn;
    int max, sockfd;
    int byte_read;
    int byte_left;

#ifdef DEBUG
    printf("am_poll enter\n");
#endif

#ifdef QUICK
    /*This  value is precomputed for quicker response times--yeah, right*/
    TCPAM_temp_fd_set = TCPAM_all_fd_set;
#else
    TCPAM_temp_fd_set= TCPAM_fd_set[REQ_CONN];
    OR_FD_SET(TCPAM_temp_fd_set, TCPAM_fd_set[REPLY_CONN]);
#endif
    max = MAX(TCPAM_max_sockfd[REQ_CONN], TCPAM_max_sockfd[REPLY_CONN]);

    if(select(max, &TCPAM_temp_fd_set, (fd_set *) 0,
	   (fd_set *) 0, &TCPAM_timeval) == -1) {
	AM_errno = errno;
	am_exit(-1);
    }

    for(i = 0; i < AM_procs; i++) {
	for(sn = 0; sn < TCPAM_NUM_CONNECTIONS; sn++) {
	    sockfd = node_sockfd[i][sn];

	    if(FD_ISSET(sockfd, &TCPAM_temp_fd_set)) {
		pp = &pkt_headers[i][sn];

		/* check if waiting for message header */
		if(message_state[i][sn] == WAIT_HEADER) {
#ifdef DEBUG
		    printf("am_poll: read header...");
#endif
		    n = sizeof(TCPAM_pkt) - hdr_bytes_read[i][sn];
		    if((m = read(sockfd, pp + hdr_bytes_read[i][sn], n)) == -1) {
          if (((errno	!= EAGAIN) && (errno != 0))|| (retry >=  MAX_RETRY))	 {
			AM_errno = errno;
      			printf("node %d am_poll: read header error", AM_myproc);
      			perror("am_poll: read header error");
 			      am_exit(-1);
        	}
          retry++;

		    } else {
retry = 0;
			hdr_bytes_read[i][sn] += m;

			if(hdr_bytes_read[i][sn] == sizeof(TCPAM_pkt)) {
			    hdr_bytes_read[i][sn] = 0;
#ifdef DEBUG
			    printf("am_poll %x %x %x %x %x %x %x %x\n", 
				   pp->tag, pp->handler, pp->calling_node, pp->arg1, 
				   pp->arg2, pp->arg3, pp->opt_arg.arg4,pp->opt_arg.arg5);
#endif
			    /* check if active message with integer arguments */
			    if (pp->tag == TCPAM_AM_INT_PKT) {
				((TCPAM_handler_4i)pp->handler)(pp->calling_node,
								pp->arg1,
								pp->arg2,
								pp->arg3,
								pp->opt_arg.arg4);

				/* check if active message with floating point arguments */
			    } else if (pp->tag == TCPAM_AM_DOUBLE_PKT) {
				((TCPAM_handler_2i1d)pp->handler)(pp->calling_node,
								  pp->opt_arg.d_arg4,
								  pp->arg1, pp->arg2);

				/* check if am_get request and do reply back */
			    } else if (pp->tag == TCPAM_GET_REQUEST_PKT) {
				am_reply_xfer(pp->calling_node,
					      pp->handler /* handler */,
					      pp->arg1 /* handler arg*/,
					      pp->arg2 /* src addr */,
					      pp->arg3 /* dst addr */,
					      pp->opt_arg.arg4 /* nbytes */);

				/* check if am_store header, and change state */
			    } else if (pp->tag == TCPAM_STORE_PKT) {
#ifdef DEBUG
				printf("reading to read data\n");
#endif
				message_state[i][sn] = WAIT_DATA;

				hp = (TCPAM_xfer_header *) &pkt_headers[i][sn];
				msg_bytes_left[i][sn] = hp->len;

				/* uh oh */
			    } else {
				/* corrupted data stream */
				printf("Error: unknown packet type\n");
				am_exit(-1);
			      }
			}
		    }

		    /* we've received hdr, waiting for data */
		} else {

		    hp = (TCPAM_xfer_header *) &pkt_headers[i][sn];
		    n = msg_bytes_left[i][sn];
		    cp = (char *) hp->addr;
#ifdef DEBUG
		    printf("am_poll: data read...\n");
#endif
		    if((m = read(sockfd, &cp[hp->len-n], n)) == -1) {
          if (((errno	!= EAGAIN) && (errno != 0)) || (retry >=  MAX_RETRY)) {	
			AM_errno = errno;					
      			printf("node %d am_poll: read data error", AM_myproc);					
      			perror("am_poll: read data error");
 			      am_exit(-1);
        	}
					retry++; 

		    }
else {
#ifdef DEBUG
		    printf("done (%d))\n", m);
#endif
retry = 0;
		    msg_bytes_left[i][sn] -= m;

		    /* invoke callback function if address is not 0 */
		    if(msg_bytes_left[i][sn] == 0) {
			message_state[i][sn] = WAIT_HEADER;

			if(hp->endfunc)
			    ((TCPAM_xfer_endfunc)hp->endfunc)
				(hp->arg2, hp->addr,
				 hp->len, (void *)(hp->arg1));
		    }
		  }
		}
	    }
	}
    }

#ifdef DEBUG
    printf("am_poll exit\n");
#endif		

    return(0);
}

/*
 * Poll request and reply networks.  Returns -1 on encountering 
 * any error and may set AM_errno, and returns 0 otherwise.
 */
int am_reply_poll()
{
    char *cp;
    TCPAM_pkt *pp;
    TCPAM_xfer_header *hp;
    fd_set TCPAM_temp_fd_set;

    int i, n, m;
    int max, sockfd;
    int byte_read, byte_left;

#ifdef DEBUG
    printf("am_reply_poll enter\n");
#endif

    TCPAM_temp_fd_set= TCPAM_fd_set[REPLY_CONN];
    max = TCPAM_max_sockfd[REPLY_CONN];

    if(select(max, &TCPAM_temp_fd_set, (fd_set *) 0,
	   (fd_set *) 0, &TCPAM_timeval) == -1) {
	AM_errno = errno;
	am_exit(-1);
    }

    for(i = 0; i < AM_procs; i++) {
	sockfd = node_sockfd[i][REPLY_CONN];

	if(FD_ISSET(sockfd, &TCPAM_temp_fd_set)) {
	    pp = &pkt_headers[i][REPLY_CONN];

	    /* check if waiting for message header */
	    if(message_state[i][REPLY_CONN] == WAIT_HEADER) {
#ifdef DEBUG
		printf("am_reply_poll: read header...");
#endif
		n = sizeof(TCPAM_pkt) - hdr_bytes_read[i][REPLY_CONN];		    
		if((m = read(sockfd, pp + hdr_bytes_read[i][REPLY_CONN], n)) == -1) {
      if (((errno	!= EAGAIN) && (errno != 0)) || (retry >=  MAX_RETRY))	 {
            AM_errno = errno;						
      			printf("node %d am_poll: read data error", AM_myproc);					
      			perror("am_poll: read data error");
 			      am_exit(-1);
        	}
					retry++; 		

		} else {
retry = 0;
		    hdr_bytes_read[i][REPLY_CONN] += m;

		    if(hdr_bytes_read[i][REPLY_CONN] == sizeof(TCPAM_pkt)) {
			hdr_bytes_read[i][REPLY_CONN] = 0;
#ifdef DEBUG
			printf("am_reply_poll %x %x %x %x %x %x %x %x\n", 
			       pp->tag, pp->handler, pp->calling_node, pp->arg1, 
			       pp->arg2, pp->arg3, pp->opt_arg.arg4,pp->opt_arg.arg5);
#endif
			/* check if active message with integer arguments */
			if (pp->tag == TCPAM_AM_INT_PKT) {
			    ((TCPAM_handler_4i)pp->handler)(pp->calling_node,
							    pp->arg1,
							    pp->arg2,
							    pp->arg3,
							    pp->opt_arg.arg4);

			    /* check if active message with floating point arguments */
			} else if (pp->tag == TCPAM_AM_DOUBLE_PKT) {
			    ((TCPAM_handler_2i1d)pp->handler)(pp->calling_node,
							      pp->opt_arg.d_arg4,
							      pp->arg1, pp->arg2);

			    /* check if am_get request and do reply back */
			} else if (pp->tag == TCPAM_GET_REQUEST_PKT) {
			    am_reply_xfer(pp->calling_node,
					  pp->handler /* handler */,
					  pp->arg1 /* handler arg*/,
					  pp->arg2 /* src addr */,
					  pp->arg3 /* dst addr */,
					  pp->opt_arg.arg4 /* nbytes */);

			    /* check if am_store header, and change state */
			} else if (pp->tag == TCPAM_STORE_PKT) {
#ifdef DEBUG
			    printf("ready to read data\n");
#endif
			    message_state[i][REPLY_CONN] = WAIT_DATA;

			    hp = (TCPAM_xfer_header *) &pkt_headers[i][REPLY_CONN];
			    msg_bytes_left[i][REPLY_CONN] = hp->len;

			    /* uh oh */
			} else {
			    /* corrupted data stream */
			printf("Error: unknown packet type\n");
			am_exit(-1);
			  }
		    }
		}

		/* we've received hdr, waiting for data */
	    } else {

		hp = (TCPAM_xfer_header *) &pkt_headers[i][REPLY_CONN];
		n = msg_bytes_left[i][REPLY_CONN];
		cp = (char *) hp->addr;
#ifdef DEBUG
		printf("am_reply_poll: data read...\n");
#endif
		if((m = read(sockfd, &cp[hp->len-n], n)) == -1) {
      if (((errno	!= EAGAIN) && (errno != 0)) || (retry >=  MAX_RETRY))	 {
		    AM_errno = errno;						
      			printf("node %d am_poll: read data error", AM_myproc);					
      			perror("am_poll: read data error");
 			      am_exit(-1);
        	}
					retry++; 


		}
else {
#ifdef DEBUG
		printf("done (%d)\n", m);
#endif
retry = 0;
		msg_bytes_left[i][REPLY_CONN] -= m;

		/* invoke callback function if address is not 0 */
		if(msg_bytes_left[i][REPLY_CONN] == 0 && hp->endfunc) {
		    message_state[i][REPLY_CONN] = WAIT_HEADER;

		    ((TCPAM_xfer_endfunc)hp->endfunc)(hp->arg2, hp->addr,
						      hp->len, (void *)(hp->arg1));
		}
	      }
	    }
	}
    }

#ifdef DEBUG
    printf("am_reply_poll exit\n");
#endif		

    return(0);
}

/*
 * while *flag != value, poll for messages.  If poll 
 * encounters any errors, am_wait returns -1 and does
 * not modify the flag.
 */ 
int am_wait(volatile int *flag, 
	    int value)
{
    while (*flag < value) {
	if(am_poll() == -1)
	    return(-1);
    }
    (*flag)-=value;
}

/*
 * poll for messages until *flag == value.  If poll encounters
 * any errors, am_poll_wait returns -1 and does not modify the
 * flag.
 */
int am_poll_wait(volatile int *flag, 
		 int value)
{
    do {
	if(am_poll() == -1)
	    return(-1);
    } while(*flag < value);
    (*flag)-=value;
}

/*
 * Sends request with 4 interger arguments. Returns -1 on
 * encountering any error and sets AM_errno, return 0 otherwise.
 */
int am_request_4(vnn_t dest,
		 handler_t fun,
		 int arg1, int arg2,
		 int arg3, int arg4)
{
    char *bp;
    TCPAM_pkt TCPAM_sbuf;

    int byte_left;
    int dest_sockfd;
    int byte_written;

    if(dest != AM_myproc) {
	bp = (char *)&TCPAM_sbuf;
	dest_sockfd = node_sockfd[dest][REQ_CONN];
	TCPAM_sbuf.tag = TCPAM_AM_INT_PKT;
	TCPAM_sbuf.calling_node = AM_myproc;
	TCPAM_sbuf.handler = (handler_t)fun;
	TCPAM_sbuf.arg1 = arg1;
	TCPAM_sbuf.arg2 = arg2;
	TCPAM_sbuf.arg3 = arg3;
	TCPAM_sbuf.opt_arg.arg4 = arg4;

	byte_left = sizeof(TCPAM_pkt);
	while(byte_left > 0) {
	    byte_written = write(dest_sockfd, bp, byte_left);
#ifdef DEBUG
	    printf("am_request_4 writes %d bytes\n", byte_written);
#endif				
	    if(byte_written > 0) {
		byte_left -= byte_written;
		bp += byte_written;

	    } else {
		if (errno != EWOULDBLOCK && errno != EAGAIN && errno != 0) {
		    AM_errno = errno;
		    am_exit(-1);
		}
		am_poll();
	    }
	}

    } else {
	/* loopback */
	((TCPAM_handler_4i)fun)(AM_myproc, 
				arg1, arg2,
				arg3, arg4);
	am_poll();		
    }

    return(0);
}

/*
 * Sends request with 2 interger arguments and 1 double argument.
 * Returns -1 on encountering any error and sets AM_errno, return 0 
 * otherwise.
 */
int am_request_fd(vnn_t dest,
		  handler_t fun,
		  double arg1,
		  int arg2,
		  int arg3)
{
    char *bp;
    TCPAM_pkt TCPAM_sbuf;

    int byte_left;
    int dest_sockfd;
    int byte_written;

    if(dest != AM_myproc) {
	bp = (char *)&TCPAM_sbuf;
	dest_sockfd = node_sockfd[dest][REQ_CONN];
	TCPAM_sbuf.tag = TCPAM_AM_DOUBLE_PKT;
	TCPAM_sbuf.calling_node = AM_myproc;
	TCPAM_sbuf.handler = (handler_t)fun;
	TCPAM_sbuf.arg1 = arg2;
	TCPAM_sbuf.arg2 = arg3;
	TCPAM_sbuf.opt_arg.d_arg4 = arg1;

	byte_left = sizeof(TCPAM_pkt);
	while(byte_left > 0) {
	    byte_written = write(dest_sockfd, bp, byte_left);
#ifdef DEBUG
	    printf("am_request_fd writes %d bytes\n", byte_written);
#endif				
	    if(byte_written > 0) {
		byte_left -= byte_written;
		bp += byte_written;

	    } else {
		if(errno != EWOULDBLOCK && errno != EAGAIN && errno != 0) {
		    AM_errno = errno;
		    am_exit(-1);
		}
		am_poll();
	    }
	}

    } else {
	/* loopback */
	((TCPAM_handler_2i1d)fun)(AM_myproc, 
				  arg1,
				  arg2,
				  arg3);
	am_poll();		
    }

    return(0);
}

/*
 * Sends a reply message with 4 integer arguments.  Returns -1 on
 * encountering any error and sets AM_errno, return 0 otherwise.
 */
int am_reply_4(vnn_t dest,
	       handler_t fun,
	       int arg1, int arg2, 
	       int arg3, int arg4)
{
    char *bp;
    TCPAM_pkt TCPAM_sbuf;

    int byte_left;
    int dest_sockfd;
    int byte_written;

    if(dest != AM_myproc) {
	bp = (char *)&TCPAM_sbuf;
	dest_sockfd = node_sockfd[dest][REPLY_CONN];
	TCPAM_sbuf.tag = TCPAM_AM_INT_PKT;
	TCPAM_sbuf.calling_node = AM_myproc;
	TCPAM_sbuf.handler = (handler_t)fun;
	TCPAM_sbuf.arg1 = arg1;
	TCPAM_sbuf.arg2 = arg2;
	TCPAM_sbuf.arg3 = arg3;
	TCPAM_sbuf.opt_arg.arg4 = arg4;

	byte_left = sizeof(TCPAM_pkt);
	while(byte_left > 0) {
	    byte_written = write(dest_sockfd, bp, byte_left);
#ifdef DEBUG
	    printf("am_reply_4 writes %d bytes\n", byte_written);
#endif							
	    if(byte_written > 0) {
		byte_left -= byte_written;
		bp += byte_written;

	    } else {
		if(errno != EWOULDBLOCK && errno != EAGAIN && errno != 0) {
		    AM_errno = errno;
		    am_exit(-1);
		}
		am_reply_poll();
	    }
	}

    } else {
	/* loopback */
	((TCPAM_handler_4i)fun)(AM_myproc, 
				arg1, arg2,
				arg3, arg4);
	am_reply_poll();
    }

    return(0);
}

/*
 * Sends a reply message with 3 integer arguments and 1 
 * double-precision floating point argument.  Returns -1 on
 * encountering any error and sets AM_errno, return 0 otherwise.
 */
int am_reply_fd(vnn_t dest,
		handler_t fun,
		double arg1,
		int arg2,
		int arg3)
{
    register char *bp;
    TCPAM_pkt TCPAM_sbuf;

    int byte_left;
    int dest_sockfd;
    int byte_written;

    if(dest != AM_myproc) {
	bp = (char *)&TCPAM_sbuf;
	dest_sockfd = node_sockfd[dest][REPLY_CONN];
	TCPAM_sbuf.tag = TCPAM_AM_DOUBLE_PKT;
	TCPAM_sbuf.calling_node = AM_myproc;
	TCPAM_sbuf.handler = (handler_t)fun;
	TCPAM_sbuf.arg1 = arg2;
	TCPAM_sbuf.arg2 = arg3;
	TCPAM_sbuf.opt_arg.d_arg4 = arg1;

	byte_left = sizeof(TCPAM_pkt);
	while(byte_left > 0) {
	    byte_written = write(dest_sockfd, bp, byte_left);
#ifdef DEBUG
	    printf("am_reply_fd writes %d bytes\n", byte_written);
#endif							
	    if(byte_written > 0) {
		byte_left -= byte_written;
		bp += byte_written;

	    } else {
		if(errno != EWOULDBLOCK && errno != EAGAIN && errno != 0) {
		    AM_errno = errno;
		    am_exit(-1);
		}
		am_reply_poll();
	    }
	}

    } else {
	/* loopback */
	((TCPAM_handler_2i1d)fun)(AM_myproc,
				  arg1,
				  arg2,
				  arg3);
	am_reply_poll();
    }

    return(0);
}

/*
 * Copy nbytes bytes of data from address lva on calling node to address
 * rva on dest node. Invoke handler with handler_arg on dest node when all
 * data are received on dest node.  Returns -1 on encountering any error and
 * sets AM_errno, returns 0 otherwise.
 */
int am_store(vnn_t dest,
	     void *lva,
	     void *rva,
	     int nbytes,
	     handler_t request_handler,
	     void *handler_arg)
{
    char *bp;
    TCPAM_xfer_header header;

    int byte_left;
    int dest_sockfd;
    int byte_written;

    if(dest != AM_myproc) {
	header.tag = TCPAM_STORE_PKT;
	header.len = nbytes;
	header.addr = rva;
	header.endfunc = request_handler;
	header.arg1 = (int)handler_arg;
	header.arg2 = AM_myproc;
	dest_sockfd = node_sockfd[dest][REQ_CONN];
	bp = (char *)&header;
		
	/* send the header */
	byte_left = sizeof(TCPAM_xfer_header);
	while(byte_left > 0) {
	    byte_written = write(dest_sockfd, bp, byte_left);
#ifdef DEBUG
	    printf("am_store writes %d bytes of header\n", byte_written);
#endif			
	    if(byte_written > 0) {
		byte_left -= byte_written;
		bp += byte_written;

	    } else {
		if(errno != EWOULDBLOCK && errno != EAGAIN && errno != 0) {
		    AM_errno = errno;
		    am_exit(-1);
		}
		am_poll();
	    }
	}
		
	/* send the buffer */
	bp = lva;		
	byte_left = nbytes;
	while(byte_left > 0) {
	    byte_written = write(dest_sockfd, bp, byte_left);
#ifdef DEBUG
	    printf("am_store %d/%d\n", byte_written, nbytes);
#endif
	    if(byte_written > 0) {
		byte_left -= byte_written;
		bp += byte_written;

	    } else {
		if(errno != EWOULDBLOCK && errno != EAGAIN && errno != 0) {
		    AM_errno = errno;
		    am_exit(-1);
		}
		am_poll();
	    }
	}		

    } else {
	/* loopback */
	memmove(rva, lva, nbytes);

	/* invoke callback function if address is not 0 */
	if(request_handler)
	    ((TCPAM_xfer_endfunc)request_handler)(AM_myproc, 
						  rva, nbytes,
						  handler_arg);
	am_poll();
    }

    return(0);
}


/* Could have used non-blocking send to implement this. Not sure it
 * is better though -- ltl
 */
int am_store_async(vnn_t dest, void *src_buf, void *dest_buf,
									 int len, handler_t request_handler,
									 void *request_handler_arg,
									 handler_t endfunc, void *endfunc_arg)
{
  am_store(dest, src_buf, dest_buf, len, request_handler,
					 request_handler_arg);
	/* invoke the handler */
	(endfunc)(dest, src_buf, len, endfunc_arg);
}

/*
 * Like am_store, but used for the reply bulk transfer for am_gets.
 * Also, am_reply_xfer only polls the reply network.  Returns -1 on 
 * encountering any errors and sets AM_errno, and 0 otherwise.
 */
int am_reply_xfer(vnn_t dest, 
		  handler_t request_handler,
		  void *handler_arg,
		  void *lva,
		  void *rva, 
		  int nbytes)
{
    char *bp;
    TCPAM_xfer_header header;

    int byte_left;
    int dest_sockfd;
    int byte_written;

    if(dest != AM_myproc) {
	header.tag = TCPAM_STORE_PKT;
	header.endfunc = request_handler;
	header.arg1 = (int)handler_arg;
	header.arg2 = AM_myproc;
	header.addr = rva;
	header.len = nbytes;
	dest_sockfd = node_sockfd[dest][REPLY_CONN];
	bp = (char *) &header;
		
	/* send the header */
	byte_left = sizeof(TCPAM_xfer_header);
	while(byte_left > 0) {
	    byte_written = write(dest_sockfd, bp, byte_left);
#ifdef DEBUG
	    printf("am_reply_xfer writes %d bytes of header\n", byte_written);
#endif			
	    if(byte_written > 0) {
		byte_left -= byte_written;
		bp += byte_written;

	    } else {
		if(errno != EWOULDBLOCK && errno != EAGAIN && errno != 0) {
		    AM_errno = errno;
		    am_exit(-1);
		}
		am_reply_poll();
	    }
	}
		
	/* send the data buffer */
	bp = lva;		
	byte_left = nbytes;

	while(byte_left > 0) {
	    byte_written = write(dest_sockfd, bp, byte_left);
#ifdef DEBUG
	    printf("am_reply_xfer writes %d bytes of buffer\n", byte_written);
#endif						
	    if(byte_written > 0) {
		byte_left -= byte_written;
		bp += byte_written;

	    } else {
		if(errno != EWOULDBLOCK && errno != EAGAIN && errno != 0) {
		    AM_errno = errno;
		    am_exit(-1);
		}
		am_reply_poll();
	    }
	}		

    } else {
	/* loopback */
	memmove(rva, lva, nbytes);

	/* invoke callback function if address is not 0 */
	if(request_handler)
	    ((TCPAM_xfer_endfunc)request_handler)(AM_myproc,
						  rva, nbytes,
						  handler_arg);
	am_reply_poll();
    }

    return(0);
}

/*
 * Retrive nbytes bytes of data from address rva on dest node to address
 * lva on calling node.  Invokes handler with handler_arg on calling node 
 * when all data are received on calling node.  Return -1 on encountering
 * any errors and sets AM_errno, and 0 otherwise.
 */
int am_get(vnn_t dest,
	   void *sva,
	   void *dva,
	   int nbytes,
	   handler_t reply_handler,
	   void *handler_arg)
{
    char *bp;
    TCPAM_pkt TCPAM_sbuf;

    int byte_left;
    int dest_sockfd;
    int byte_written;

    if(dest != AM_myproc) {
	bp = (char *) &TCPAM_sbuf;
	dest_sockfd = node_sockfd[dest][REQ_CONN];
	TCPAM_sbuf.tag = TCPAM_GET_REQUEST_PKT;
	TCPAM_sbuf.calling_node = AM_myproc;
	TCPAM_sbuf.handler = reply_handler;
	TCPAM_sbuf.arg1 = (int)handler_arg;
	TCPAM_sbuf.arg2 = (int)sva;
	TCPAM_sbuf.arg3 = (int)dva;
	TCPAM_sbuf.opt_arg.arg4 = nbytes;
	byte_left = sizeof(TCPAM_pkt);

	while(byte_left > 0) {
	    byte_written = write(dest_sockfd, bp, byte_left);
#ifdef DEBUG
	    printf("am_get writes %d bytes\n", byte_written);
#endif				
	    if(byte_written > 0) {
		byte_left -= byte_written;
		bp += byte_written;

	    } else {
		if(errno != EWOULDBLOCK && errno != EAGAIN && errno != 0) {
		    AM_errno = errno;
		    am_exit(-1);
		}
		am_reply_poll();
	    }
	}

    } else {
	/* loopback */
	memmove(dva, sva, nbytes);

	/* invoke callback function if address is not 0 */
	if(reply_handler)
	    ((TCPAM_xfer_endfunc)reply_handler)(AM_myproc, 
						dva, nbytes,
						handler_arg);
	am_reply_poll();		
    }

    return(0);
}

/*
 * am_isalive returns TRUE if communications via GAM
 * to the specified vnn are operating "normally".  For
 * TCPAM this means that we can do a select() on the fd's
 * associated with the sockets to that specified VNN, and
 * select returns normally.
 */
int am_isalive(vnn_t vnn)
{
    fd_set fs;
    char msg_byte;

#if 0
    FD_SET(node_sockfd[vnn][REQ_CONN], &fs);
#endif

    if(recv(node_sockfd[vnn][REQ_CONN],&msg_byte,1,MSG_PEEK)==-1)
      if (errno == EIO) return (0);

#if 0
    if(select(1, &fs, (fd_set *) 0, (fd_set *) 0, &TCPAM_timeval) == -1)
	return(0);
    FD_SET(node_sockfd[vnn][REPLY_CONN], &fs);
#endif

    if(recv(node_sockfd[vnn][REPLY_CONN],&msg_byte,1,MSG_PEEK)==-1)
      if (errno == EIO) return (0);
#if 0
    if(select(1, &fs, (fd_set *) 0, (fd_set *) 0, &TCPAM_timeval) == -1)
	return(0);
#endif

    return(1);
}

/*
 * Returns virtual node number 
 */
int am_my_proc()
{
    return AM_myproc;
}

/*
 * Returns the number of nodes
 */
int am_procs()
{
    return AM_procs;	
}



/*
 * Returns AM_errno
 */
int am_errno()
{
    return(AM_errno);
}

/* Return maximum buffer size in bytes for put and get */
int am_max_size()
{
    return (SB_MAX/2);
}
