/*									tab:4
 *
 * malloc_int.h - Internal defs for spread malloc.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Andrea Dusseau
 * Version:			4
 * Creation Date:	??
 * Filename:		malloc_int.h
 * History:
 *	ACD	3	Mon Jun 21 13:13:38 1993
 *		Major clean-up.  Restructed so call CMPE_set_iu_heap_ptr from nodes
 *              instead of going through host.
 * 	TvE	2	Thu Dec 24 03:23:13 PST 1992
 *		moved tp malloc_int.h
 * 	TvE	1	Tue Sep 29 12:47:46 PDT 1992
 *		changed include file names
 */

#ifndef __MALLOC_INT_H__
#define __MALLOC_INT_H__

/*	
 * Defines things that are shared across the procedures that
 * do dynamic storage allocation (spread_malloc, spread_free, etc) but aren't
 * visible to users of the storage allocator.
 */

#define MAX_BLOCKS(i) (int)(((i) + PROCS - 1) / PROCS)

#define ERROR (-1)
#define ALLOCATED (-2)
#define NEEDSBRK (-3)

#define SPREAD_MALLOC_PROC 0

typedef struct gmalloc_type {
  volatile int flag;
  int result;
  void *ptr;
  unsigned sizetobrk;
} gmalloc_t;


/* --------------------------- Procedure Prototypes ------------------------ */
/*
 * The routines for big memory allocations
 */
void gmalloc_big_handler(int proc, unsigned bytesNeeded, gmalloc_t *info);
void morebigmem_handler(int RegionSize, void *newRegion, gmalloc_t *info);
void trytoallocbig();

/*
 * The routines for binned memory allocations
 */
void gmalloc_bin_handler(int proc, unsigned size, int index, gmalloc_t *info);
void *trytobin(unsigned bytesNeeded, int index);
void morebinmem_handler(int proc, void *newRegion, gmalloc_t *info);

/*
 * Miscellaneous routines
 */
void *sc_malloc(unsigned num_elems, unsigned bytes_per_elem);
void spread_free_handler(void *blockPtr);
void gmalloc_reply_handler(gmalloc_t *info, int result, void *ptr, int size);
void *MemChunkAlloc(unsigned size);


/* ------------------ Definitions for Local Memory Allocation ------------- */
/*
 * WARNING: several of the macros in this file expect both integers and
 * pointers to be 32 bits (4 bytes) long.
 */

/*
 * This storage allocator consists of two indepedent allocators,
 * one used for binned objects and the other used for large objects.
 * Objects <= BIN_SIZE are automatically binned, and all objects
 * larger than BIN_SIZE are allocated with the large object
 * allocator. 
 *
 * Both allocators deal with "blocks" of storage, which correspond
 * roughly to the areas that users request and free.  Each block of
 * storage consists of an administrative area (usually a single word)
 * followed by the useable area.  All allocations are done in even
 * numbers of words, which means that clients may occasionally get more
 * bytes than they asked for.  Malloc returns the address of the word
 * just after the administrative area.  The first word of the administrative
 * area is called the administrative word;  it allows us to keep trackk of
 * which blocks are in use and which are free, and also to keep track of
 * the sizes of blocks so clients don't have to know how large a block is
 * when they free it.  The low-order bit of the administrative word
 * indicates whether or not the block is free.  The next-to-low-order bit
 * of the administrative word is used to indicate that this is a "dummy
 * block":  for blocks on the un-binned list, this means that the block
 * marks an area not belonging to the allocator;  for blocks returned to
 * the "free" procedure, this means that the block must go back to the
 * binned allocator.  The rest of the word tells how large the block is
 * (in bytes including the administrative area).  Using the low-order bits
 * for flags works because all blocks that we allocate are always multiples
 * of 4 bytes in length.  There is one exception to this usage of the
 * administrative word, which occurs for free binned objects.  In this
 * case the word is a link to the next free binned object instead of a
 * size.  If tracing is enabled, then the administrative area also contains
 * a couple of other fields in addition to the administrative word.  The
 * macros below define the format of the administrative area and how to
 * access it.
 */

#define Address void *
typedef double AdminInfo;

#define GET_ADMIN(blockPtr)	(*(int *)(AdminInfo *) (blockPtr))
#define SET_ADMIN(blockPtr, value) *((int *)(AdminInfo *) (blockPtr)) = (int) (value)


#define USE_BIT		1
#define DUMMY_BIT	2
#define FLAG_BITS	3
#define ADMIN_BITS(admin)	((admin) & FLAG_BITS)
#define IS_IN_USE(admin)	((admin) & USE_BIT)
#define IS_DUMMY(admin)		(((admin) & FLAG_BITS) == (USE_BIT|DUMMY_BIT))
#define MARK_DUMMY(admin)	((admin) | (USE_BIT|DUMMY_BIT))
#define MARK_IN_USE(admin)	((admin) | USE_BIT)
#define MARK_FREE(admin)	((admin) & ~FLAG_BITS)
#define SIZE(admin)		((admin) & ~FLAG_BITS)


/*
 * ----------------------------------------------------------------------------
 *
 *	Binned-object allocator: it keeps a separate linked free list
 *	for each size of object less than SMALL_SIZE bytes in size and
 *	a free list for all objects less than BIN_SIZE that have been 
 *	specified as being binned by a call to Mem_Bin.  The blocks on
 *	each list are linked together via their administrative words
 *	terminated by a NULL pointer.  Allocation and freeing are done in
 *	a stack-like manner from the front of the free lists.
 *
 *	The following definitions are related:  if you change one, you'll
 *	have to change several:
 *
 *	GRANULARITY: The difference in size between succesive buckets.  This
 *		is determined by data alignment restrictions and should be
 *		a power of 2 to allow for shifting to map from size to bucket.
 *	SIZE_SHIFT: The shift distance to convert between size and bucket.
 *	SMALL_SIZE:		largest blocksize (total including
 *				administrative word) managed by default by
 *				the binned-object allocator.
 *	BIN_SIZE:		largest possible that can be binned.
 *
 *	These can be computed from the above constants.
 *
 *	SMALL_BUCKETS:		number of free lists corresponding to
 *				sizes <= SMALL_SIZE.
 *	BIN_BUCKETS:		total number of binned free lists.
 *	BYTES_TO_BLOCKSIZE:	how many bytes a block must contain (total
 *				including admin. word) to satisfy a user
				request in bytes.
 *	BLOCKSIZE_TO_INDEX:	which free list to use for blocks of a given
 *				size (total including administrative word).
 *	INDEX_TO_BLOCKSIZE: the (maximum) size corresponding to a bucket.
 *		This will include the size of the administrative word.
 *
 *	This is independent.
 *
 *	BLOCKS_AT_ONCE:		when allocating a new region to hold blocks of
 *				a given size, how many blocks worth should
 *				be allocated at once.
 *
 * ----------------------------------------------------------------------------
 */

/*
 * 8 byte granularity to handle SPUR and SPARC.
 */
#define GRANULARITY			8
#define SIZE_SHIFT			3

#define BIN_SIZE			271    
#define BIN_BUCKETS			((BIN_SIZE+GRANULARITY-1)/GRANULARITY) 
#define MASK				(GRANULARITY-1)
#define BYTES_TO_BLOCKSIZE(bytes) ((bytes+MASK+sizeof(AdminInfo)) & (~MASK))
#define BLOCKSIZE_TO_INDEX(size)  ((size)>>SIZE_SHIFT)
#define INDEX_TO_BLOCKSIZE(index) ((index)<<SIZE_SHIFT)

#define BLOCKS_AT_ONCE                          32

/*
 * The following array holds pointers to the first free blocks of each size.
 * Bucket i holds blocks whose total length is 4*i bytes (including the
 * administrative info).  Blocks from bucket i are used to satisfy user
 * requests ranging from (4*i)-sizeof(AdminInfo) bytes up to (4*i)-1 bytes.  
 * Buckets 0 and 1 are never used, since they correspond to blocks too small
 * to hold any information for the user.  If the list head is NOBIN, it
 * means that this size isn't binned:  use the large block allocator.
 */

extern Address	memFreeLists[BIN_BUCKETS];
#define NOBIN	((Address) -1)

/*
 * ----------------------------------------------------------------------------
 *	Large-object allocator:  used for blocks that aren't binned.
 *	There should be relatively few allocations made by the
 *	large-object allocator.  Since all the blocks it allocates are
 *	relatively large, fragmentation should be less than it would be
 *	if it allocated small blocks too. All of the blocks, in use or
 *	not, are kept in a single linked list using their administrative
 *	words.  The address of the next block in the list is found by
 *	adding the size of the current block to the address of its
 *	administrative word.  Things are complicated slightly because the
 *	storage area managed by this allocator may be discontiguous:  there
 *	could be several large regions that it manages, separated by gaps
 *	that are managed by some other storage allocator (e.g. the
 *	binned-object allocator).  In this case, the gaps between regions
 *	are handled by making the gaps look like allocated blocks, with an
 *	administrative word at the end of one region that causes a skip to
 *	the beginning of the next region.  These blocks are called "dummy
 *	blocks", and have special flag bits in their administrative words.
 *	All the blocks within a region are linked together in increasing
 *	order of address.  However, the different regions may appear in any
 *	order on the list.
 * ----------------------------------------------------------------------------
 */

extern Address memFirst;		/* Pointer to first block of memory
					 * in the un-binned pool. */
extern Address memLast;			/* Points to dummy block at the end
					 * of the memory list, which always
					 * has zero size. */
extern Address memCurrentPtr;		/* Next block to consider allocating.
					 * Rotates circularly through free
					 * list. */
extern int memLargestFree;		/* This holds the size of the largest
					 * free block that's been seen on the
					 * free list.  It's updated as the
					 * list is traversed.  */
extern int memBytesFreed;		/* Sum of unbinned bytes freed since
					 * the last time we started scanning
					 * the memory list. */

/*
 * Minimum size of new regions requested by large-object allocator
 * when storage runs out:
 */

#define MIN_REGION_SIZE 2048
#endif
