/*                                                                      tab:8
 *
 * atomic.sc - Atomic operations.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:             Andrea Dusseau
 * Version:            8         
 * Creation Date:      Thu Oct 22 15:49:07 PDT 1992
 * Filename:           atomic.sc
 * History:				
 *	ssl	1	Wed Jan 19 11:44:07 1994
 *		removed rpc
 *	ssl	8	Tue Nov 16 12:09:22 1993
 *		Replaced global tags with macros.
 *	ssl	7	Thu Oct 28 22:02:45 1993
 *		Include am.h instead of cm5.h
 *	LTL	6	Mon Aug 16 15:17:42 1993
 *		Fixed it up for version 112 to support CMMD
 *	SL	5	Wed Jan 20 11:10:14 1993
 *		        Added atomic and rpc long routines
 *	SL	4	Tue Jan 19 14:01:21 1993
 *		        Added atomic synchronization and rpc routines
 *  ACD 3   Thu Jan 14 15:42:03 PST 1993
 *              Removed "Static" qualifier from atomic_return_x_rh definitions
 *  ACD     2       Tue Jan  5 14:50:58 PST 1993
 *              Moved definitions for atomic, atomic_x from atomic_inline.h
 */

#include <stdio.h>
/*#include "am.h"*/
#define ATOMIC_INTERNAL_DEFINE

#include <split-c/split-c.h>
#include <split-c/atomic.h>

mutex_t	Atomic_mutex;
mutex_t	AtomicFunc_mutex;

void atomic_init(void)
{
#ifndef SC_SEP_REQ_RPL
#ifndef SC_AM_ATOMIC
#ifdef  SC_MSG_THR
	if (mutex_init(&Atomic_mutex, USYNC_THREAD, NULL)) {
		perror("mutex_init of Atomic_mutex");
		exit(-1);
	}
	if (mutex_init(&AtomicFunc_mutex, USYNC_THREAD, NULL)) {
		perror("mutex_init of AtomicFunc_mutex");
		exit(-1);
	}
#endif
#endif
#endif
}


static Counter atomic_sent = 0;
static atomic_t return_val;

inline __incr_atomic_ctr()
{
	__splitc_lock(&Atomic_mutex);
	atomic_sent++;
	__splitc_unlock(&Atomic_mutex);
}

inline __decr_atomic_ctr()
{
	__splitc_lock(&Atomic_mutex);
	atomic_sent--;
	__splitc_unlock(&Atomic_mutex);
}

/* ..........................................
 *
 * Initialization routine for long operations
 *
 * ..........................................
 */

/* ...........................................
 *
 * Handlers for synchronized atomic operations
 *
 * ...........................................
 */

void __atomic_nosync_handler(void *token, void (*fun)(), void *lptr, int arg2,
									int arg3, int arg4)
{
	__splitc_lock(&AtomicFunc_mutex);
	(*fun)(lptr, arg2, arg3, arg4);
	__splitc_unlock(&AtomicFunc_mutex);
	
	AM_Reply0(token, __NULL_RH);
}

void __atomic_handler(void *token, void (*fun)(), void *lptr, int arg2,
							 int arg3)
{
	__splitc_lock(&AtomicFunc_mutex);
	(*fun)(lptr, arg2, arg3);
	__splitc_unlock(&AtomicFunc_mutex);
	
	__decr_atomic_ctr();
	AM_Reply0(token, __NULL_RH);
}

void __atomic_long_handler(void *token, void (*fun)(), void *lptr, int arg1,
								  int arg2, int arg3, int arg4)
{
	__splitc_lock(&AtomicFunc_mutex);
	(*fun)(lptr, arg1, arg2, arg3, arg4);
	__splitc_unlock(&AtomicFunc_mutex);
	
	__decr_atomic_ctr();
	AM_Reply0(token, __NULL_RH);
}

void __atomic_val_handler(void *token, void (*fun)(), void *lptr, int arg2,
						  int arg3)
{
	/* This function has to call atomic_return_i or _d, which replies */
	__splitc_lock(&AtomicFunc_mutex);
	(*fun)(token, lptr, arg2, arg3);
	__splitc_unlock(&AtomicFunc_mutex);
}

/* ...........................................
 *
 * Handlers for replies from atomic operations
 *
 * ...........................................
 */

void __atomic_i_reply_handler(void *token, int val)
{
	return_val.value.v_int = val;
	return_val.flag = 1;
}

void __atomic_d_reply_handler(void *token, double val)
{
	return_val.value.v_double = val;
	return_val.flag = 1;
}

/* .............................................
 *
 * Building blocks for general atomic operations
 *
 * .............................................
 */

void atomic_no_sync(void (*fun)(), ___V_GPTR gptr, int arg2, int arg3,
					int arg4)
INLINE({
	AM_Request5(__sc_endpoint_src, toproc(gptr), __ATOMIC_NOSYNC_H, fun, tolocal(gptr),
				arg2, arg3, arg4);
})

void atomic(void (*fun)(), ___V_GPTR gptr, int arg3, int arg4)
INLINE({
	AM_Request4(__sc_endpoint_src, toproc(gptr), __ATOMIC_H, (int)fun,
				(int)tolocal(gptr), arg3, arg4);
	__incr_atomic_ctr();
})

void atomic_long(void (*fun)(), ___V_GPTR gptr, int arg1, int arg2,
				 int arg3, int arg4)
INLINE({

	AM_Request6(__sc_endpoint_src, toproc(gptr), __ATOMIC_LONG_H, fun, tolocal(gptr),
				arg1, arg2, arg3, arg4);
	__incr_atomic_ctr();
})

int atomic_i(void (*fun)(), ___V_GPTR gptr, int arg3, int arg4)
INLINE({
	return_val.flag = 0;

	AM_Request4(__sc_endpoint_src, toproc(gptr), __ATOMIC_VAL_H, (int)fun,
				(int)tolocal(gptr), arg3, arg4);
	while (return_val.flag == 0) {
		splitc_poll();
	}
	return return_val.value.v_int;
})

double atomic_d(void (*fun)(), ___V_GPTR gptr, int arg3, int arg4)
INLINE({
	return_val.flag = 0;

	AM_Request4(__sc_endpoint_src, toproc(gptr), __ATOMIC_VAL_H, (int)fun,
				(int)tolocal(gptr), arg3, arg4);	
	while (return_val.flag == 0) {
		splitc_poll();
	}
	return return_val.value.v_double;
})

/* .................................................
 *
 * Handlers for canned atomic operations on integers
 *
 * .................................................
 */

void cmp_and_swap_atomic(int proc, short *loc, int test, int new)
{
	if(*loc == test) {
		*loc = new;
		atomic_return_i(1);
	}
	atomic_return_i(0);
}

void test_and_set_atomic(int proc, int *loc)
{
	int tmp = *loc;
	*loc = 1;
	atomic_return_i(tmp);
}

void exchange_atomic(int proc, int *loc, int new_val)
{
	int tmp = *loc;
	*loc = new_val;
	atomic_return_i(tmp);
}

void fetch_and_add_atomic(int proc, int *loc, int new_val)
{
	int tmp = *loc;
	*loc += new_val;
	atomic_return_i(tmp);
}

/* .........................................
 *
 * Atomic and rpc synchronization primitives
 *
 * .........................................
 */

/* Wait until all calls to atomic from all processors have completed */
void all_atomic_sync()
{
	int res;
	extern int ___i_reduce(int diff);

	barrier();
	do {
		res = ___i_reduce(atomic_sent);
	} while(res != 0);

	__splitc_lock(&Atomic_mutex);
	atomic_sent = 0;
	__splitc_unlock(&Atomic_mutex);
}

