/* 									tab:8
 *
 * bulk.h - Global memory bulk access.
 *          Implementation dependent definitions and declarations.
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:		Andrea Dusseau
 * Version:		43
 * Creation Date:	Thu Oct  8 14:16:42 PDT 1992
 * Filename:		bulk.h
 * History:
 *	LTL	16	Sun Jul 10 15:05:54 1994
 *		Modified for TCPAM
 *	ssl	9	Sat Nov 27 22:58:42 1993
 *		Included __get* __put* __write from memory.[ch]. Included split-c.h
 *	ssl	8	Tue Nov 16 11:56:20 1993
 *		Replaced global tags with macros.
 *	AK	7	Mon Apr 12 15:13:03 1993
 *		Changed bulk_store decl to "inline". To be consistent
 *		with split-c.h
 *	AK	4	Sat Jan 23 02:38:09 1993
 *		Made bulk_store independant of __store.
 * 	TvE	2	Thu Dec 24 04:30:31 PST 1992
 *		Rewrite. Now entirely based on PUTs, GETs and STOREs
 *	AK	1	Sat Nov 28 21:17:06 1992
 *		Changed include file name from cmam2-4
 *	DEC	1	Sat Oct 24 01:21:37 1992
 *		Changed store counters to recvBytes and sendBytes to be
 *      	compatible with split-c_int of non-bulk ops.
 */

#ifndef __BULK_INT_H__
#define __BULK_INT_H__

#include <split-c/split-c.h>

/* ....................
 *
 * Bulk PUTs and WRITEs
 *
 * ....................
 */


extern void __GetComplete(int src_node, void *buf, int len, void *info);
extern void __PutCtrCompleteAtSource(int src_node, Counter *ctr);
extern void __bulk_put_handler(int dest, void *buf, int len, void *cnt);

/* .........
 *
 * Bulk GETs
 *
 * .........
 */

/* Explicit counter */
EXTERN inline void bulk_get_ctr(void *lptr, ___V_GPTR gptr,
				int len, Counter *ctr)
INLINE({
	__get_ctr(lptr, gptr, len, ctr);
})

/* Implicit GET counter */
EXTERN inline void bulk_get(void *lptr, ___V_GPTR gptr, int len)
INLINE({
	__get_ctr(lptr, gptr, len, &numGetOuts);
})

/* Blocking */
EXTERN inline void bulk_read(void *lptr, ___V_GPTR gptr, int len)
INLINE({
	Counter ctr = 0;
	__get_ctr(lptr, gptr, len, &ctr);
	sync_ctr(&ctr);
})

/* GET bulk with explicit counter */
EXTERN inline void __get_ctr(void *lPtr, ___V_GPTR gPtr,
			     int len, Counter *ctr)
INLINE({
	int segId;
	int pid    = toproc(gPtr);
	void *addr = tolocal(gPtr);
	int i;

	if (pid == MYPROC) {
		memcpy(lPtr,addr, len);
	} 
	else {
	  (*ctr)++;
		if (len <= AM_MAX_PKT) {
      am_get(pid, addr, lPtr, len, __GetComplete, (void *)ctr);
		}
		else {
			for (i = AM_MAX_PKT; i < len; i += AM_MAX_PKT) {
				am_get(pid, addr, lPtr, AM_MAX_PKT, __GetComplete, 0);
				(char *)addr += AM_MAX_PKT;
				(char *)lPtr += AM_MAX_PKT;
		  }
			i -= AM_MAX_PKT;
			am_get(pid, addr, lPtr, len - i, __GetComplete, (void *)ctr);
	  }
  }
})

EXTERN inline void __get(void *lPtr, ___V_GPTR gPtr, int len)
INLINE({ __get_ctr(lPtr, gPtr, len, &numGetOuts); })
	

/* .........
 *
 * Bulk PUTs
 *
 * .........
 */

EXTERN inline void bulk_put_ctr(___V_GPTR gptr, void *lptr,
				int len, Counter *ctr)
INLINE({
	__put_ctr(gptr, lptr, len, ctr);
})

EXTERN inline void bulk_put(___V_GPTR gptr, void *lptr, int len)
INLINE({
	__put(gptr, lptr, len);
})

EXTERN inline void bulk_write(___V_GPTR gptr, void *lptr, int len)
INLINE({
	__write(gptr, lptr, len);
})

EXTERN inline void __put_ctr(___V_GPTR gPtr, void *lPtr, int len, Counter *ctr)
INLINE({
	Processor pid = toproc(gPtr);
	void *addr    = tolocal(gPtr);
	int i;

	if(pid == MYPROC) {
		memcpy(addr,lPtr, len);
	} else {
	  (*ctr)++;
		if (len <= AM_MAX_PKT) {
      am_store_async(pid, lPtr, addr, len, __bulk_put_handler, (void *)ctr);
		}
		else {
			for (i = AM_MAX_PKT; i < len; i += AM_MAX_PKT) {
				am_store_async(pid, lPtr, addr, AM_MAX_PKT, __bulk_put_handler, 0);
				(char *)addr += AM_MAX_PKT;
				(char *)lPtr += AM_MAX_PKT;
		  }
			i -= AM_MAX_PKT;
			am_store_async(pid, lPtr, addr,len - i, __bulk_put_handler, (void *)ctr);
  	}
	}
})

EXTERN inline void __put(___V_GPTR gPtr, void *lPtr, int len)
INLINE({
	__put_ctr(gPtr,lPtr,len,&numPutOuts);
})

EXTERN inline void __write(___V_GPTR gPtr, void *lPtr, int len)
INLINE({
	__put_ctr(gPtr,lPtr,len,&numPutOuts);
	sync_ctr(&numPutOuts);
})


/* ...........
 *
 * Bulk STOREs
 *
 * ...........
 */


EXTERN void bulk_store(___V_GPTR gptr, void *lptr, int len);



#endif











