/*									tab:8
 *
 * store.h - Global memory accesses without acks (store)
 *           Implementation dependent definitions and declarations.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Arvind Krishnamurthy
 * Version:		68
 * Creation Date:	Tue Sep 22 23:54:02 1992
 * Filename:		store.h
 * History:
 *	ssl	49	Mon Mar 21 09:48:32 1994
 *		store_sync(_ctr) for HPAM can use poll but PMAM cannot
 *	ssl	48	Sun Mar 20 14:23:29 1994
 *		store_sync(_ctr) on hps can now use poll
 *	ssl	47	Sun Feb 27 15:59:51 1994
 *		added is_store_sync_ctr
 *	ssl	46	Fri Mar 11 15:22:17 1994
 *		store_sync must call am_wait on HPAM. am_poll does not retransmit.
 *      same for store_sync_ctr
 *	ssl	45	Wed Jan 19 00:05:27 1994
 *		changed include of am.h to split-c.h
 *	ssl	44	Sun Dec  5 23:07:15 1993
 *		am and handlers now use unsigned instead of long long as arguments.
 *      doubles will be passed in fp registers and long longs will be dword
 *      aligned.
 *	ssl	43	Tue Nov 16 12:04:48 1993
 *		Replaced global tags with macros.
 *	ssl	42	Thu Oct 28 22:00:54 1993
 *		Include am.h instead of cm5.h
 *	LTL	41	Mon Aug 16 15:55:41 1993
 *		Renamed some am_* macros for version 112 to support CMMD
 *	SL	40	Thu May 13 16:55:25 1993
 *		Added SC_PROFILE information
 *  AK  39  Wed May 12 23:17:37 1993
 *      Fixed reset counters.
 *	AK	38	Tue Mar 23 12:32:34 1993
 *		Fixed a bug in __f_store. Problem is the sparc calling
 *		convention that forces us to use doubles/ints.
 * 	TvE	32	Sat Dec 19 00:15:28 PST 1992
 *		Started major clean-up and overhaul.
 *		Extracted from split-c_int.h
 *	AK	1	Sat Nov 28 21:17:44 1992
 *		changed include file name from cmam2-4
 * 	TvE	31	Tue Sep 29 12:47:46 PDT 1992
 *		changed include file names
 */

#ifndef __STORE_H
#define __STORE_H

#include <split-c/split-c.h>

/* .....................
 *
 * Profiling information
 *
 * .....................
 */

#ifdef SC_PROFILE
#include <split-c/profile.h>
#endif

/* .....................
 *
 * Keep track of STORESs
 *
 * .....................
 */

/* A local variable which is incremented and then we wait for it to go to zero.
   Used as the implicit flag in all non-blocking operations */
extern Counter recvBytes;
extern Counter sentBytes;

/* Forget about all stores */
EXTERN inline void reset_store_sync(void)
INLINE({
	sentBytes -= recvBytes;
	recvBytes = 0;
})

#ifdef PMAM
EXTERN inline void store_sync(int expecting)
INLINE({
	am_wait(&recvBytes, expecting); 
	sentBytes -= expecting;
})
#elif SC_PROFILE /* Waiting for synchronization */
/* Wait for all stores *to* this processor to complete */
EXTERN inline void store_sync(int expecting)
	INLINE({
		while(recvBytes < expecting) {
			sc_prof_data[PROF_SYNC]++;
			am_poll(); 
		}
		sentBytes -= expecting;
		recvBytes -= expecting;
	})
#else
	EXTERN inline void store_sync(int expecting)
	INLINE({
		while(recvBytes < expecting) {
			am_poll(); 
		}
		sentBytes -= expecting;
		recvBytes -= expecting;
	})
#endif

#ifdef PMAM
EXTERN inline void store_sync_ctr(int expecting,
								  Counter *rBytes, Counter *sBytes)
INLINE({
	am_wait(rBytes, expecting); 
	*sBytes -= expecting;
})
#else
EXTERN inline void store_sync_ctr(int expecting,
								  Counter *rBytes, Counter *sBytes)
INLINE({
	while(*rBytes < expecting) {
		am_poll(); 
	}
	*sBytes -= expecting;
	*rBytes -= expecting;
})
#endif

/* Check whether all stores *to* this processor have completed */
EXTERN inline int is_store_sync(int expecting)
INLINE({
	am_poll(); 
	return recvBytes >= expecting;
})

EXTERN inline int is_store_sync_ctr(int expecting,
									Counter *rbytes, Counter *sbytes)
INLINE({
	am_poll(); 
	return *rbytes >= expecting;
})

#ifdef DEPRECATED
EXTERN inline int store_rcvd_bytes(void)
INLINE({
	return recvBytes;
})
#endif

/* ............................
 *
 * STOREs with explicit counter
 *
 * ............................
 */

/* Handler on the remote end to store value */
extern void __c_store_handler (int src_node, char *addr, int data,
															 Counter *ctr);
extern void __sh_store_handler(int src_node, short *addr, int data,
															 Counter *ctr);
extern void __i_store_handler (int src_node, int *addr, int data,
															 Counter *ctr, int *adr2);
extern void __f_store_handler (int src_node, float *addr, unsigned word0,
															 Counter *ctr);
extern void __d_store_handler (int src_node, double *addr,    unsigned word0,
							   unsigned word1,    Counter *ctr);
#ifdef __GNUC__
extern void __ll_store_handler(int src_node, unsigned *addr, unsigned word0,
							   unsigned word1,    Counter *ctr);
#endif

/* STORE character */
EXTERN inline void __c_store_ctr(___V_GPTR gPtr, char val,
				 Counter *rbytes, Counter *sbytes)
INLINE({
	Processor pid = toproc(gPtr);
	char *addr    = tolocal(gPtr);

	if(pid == MYPROC) {
		*sbytes += sizeof(char);
		*addr = val;
		*rbytes += sizeof(char);
	} else {
		*sbytes += sizeof(char);
		am_request_3_one_way(pid, __c_store_handler, addr, val, rbytes);
	}
})

/* STORE short */
EXTERN inline void __sh_store_ctr(___V_GPTR gPtr, short val,
				  Counter *rbytes, Counter *sbytes)
INLINE({
	Processor pid = toproc(gPtr);
	short *addr   = tolocal(gPtr);

	if(pid == MYPROC) {
		*sbytes += sizeof(short);
		*addr = val;
		*rbytes += sizeof(short);
	} else {
		*sbytes += sizeof(short);
		am_request_3_one_way(pid, __sh_store_handler, addr, val, rbytes);
	}
})

/* STORE int */
EXTERN inline void __i_store_ctr(___V_GPTR gPtr, int val,
				 Counter *rbytes, Counter *sbytes)
INLINE({
	Processor pid = toproc(gPtr);
	int *addr     = tolocal(gPtr);

	if(pid == MYPROC) {
		*sbytes += sizeof(int);
		*addr = val;
		*rbytes += sizeof(int);
	} else {
		*sbytes += sizeof(int);
		am_request_4_one_way(pid, __i_store_handler, addr, val, rbytes, addr);
	}
})

/* STORE float */
EXTERN inline void __f_store_ctr(___V_GPTR gPtr, float val,
				 Counter *rbytes, Counter *sbytes)
INLINE({
	Processor pid = toproc(gPtr);
	float *addr   = tolocal(gPtr);
	unsigned *word0 = (unsigned*) &val;

	if(pid == MYPROC) {
		*sbytes += sizeof(float);
		*addr = val;
		*rbytes += sizeof(float);
	} else {
		*sbytes += sizeof(float);
		/* pass unsigned so that it won't pass in fp registers */
		am_request_3_one_way(pid, __f_store_handler, addr, *word0, rbytes);
	}
})

/* STORE double */
EXTERN inline void __d_store_ctr(___V_GPTR gPtr, double val,
				 Counter *rbytes, Counter *sbytes)
INLINE({
	Processor pid = toproc(gPtr);
	double *addr   = tolocal(gPtr);
	unsigned *word0 = (unsigned*)&val;

	if(pid == MYPROC) {
		*sbytes += sizeof(double);
		*addr = val;
		*rbytes += sizeof(double);
	} else {
		*sbytes += sizeof(double);
		/* pass unsigned so it will not pass in fp regs */
		am_request_4_one_way(pid, __d_store_handler, addr, *word0, *(word0+1), rbytes);
	}
})

#ifdef __GNUC__
/* STORE long long */
EXTERN inline void __ll_store_ctr(___V_GPTR gPtr, long long val,
				  Counter *rbytes, Counter *sbytes)
INLINE({
	Processor pid = toproc(gPtr);
	long long *addr   = tolocal(gPtr);
	unsigned *word0 = (unsigned*) &val;

	if(pid == MYPROC) {
		*sbytes += sizeof(long long);
		*addr = val;
		*rbytes += sizeof(long long);
	} else {
		*sbytes += sizeof(long long);
		/* pass as two unsigned so long long won't be dword aligned */
		am_request_4_one_way(pid, __ll_store_handler, addr, *word0,
					*(word0+1), rbytes);
	}
})
#endif

/* STORE 4 byte structure */
EXTERN inline void __struct_i_store_ctr(___V_GPTR gPtr, ___Struct_i val,
					Counter *rbytes, Counter *sbytes)
INLINE({
	Processor pid = toproc(gPtr);
	int *addr     = tolocal(gPtr);

	if(pid == MYPROC) {
		*sbytes += sizeof(int);
		*addr = val.data;
		*rbytes += sizeof(int);
	} else {
		*sbytes += sizeof(int);
		am_request_4_one_way(pid, __i_store_handler, (int)addr, (int)val.data,
								 (int)rbytes, (int)addr);
	}
})

#ifdef __GNUC__
#ifdef PAM
/* STORE 8 byte structure */
EXTERN inline void __struct_d_store_ctr(___V_GPTR gPtr, ___Struct_d val,
				  Counter *rbytes, Counter *sbytes)
INLINE({
  union {
			long long full;
			struct {
				long lo;
				long hi;
			} half;
	} u;

	Processor pid   = toproc(gPtr);
	long long *addr = tolocal(gPtr);

	if(pid == MYPROC) {
		*sbytes += sizeof(long long);
		*addr = val.data;
		*rbytes += sizeof(long long);
	} else {
		u.full = val.data;	
		*sbytes += sizeof(long long);
		am_request_4_one_way(pid, __ll_store_handler, addr, u.half.lo,
								 u.half.hi, rbytes);
	}
})

#else

/* STORE 8 byte structure */
EXTERN inline void __struct_d_store_ctr(___V_GPTR gPtr, ___Struct_d val,
				  Counter *rbytes, Counter *sbytes)
INLINE({
	Processor pid   = toproc(gPtr);
	long long *addr = tolocal(gPtr);

	if(pid == MYPROC) {
		*sbytes += sizeof(long long);
		*addr = val.data;
		*rbytes += sizeof(long long);
	} else {
		*sbytes += sizeof(long long);
		am_request_3_one_way(pid, __ll_store_handler, addr, val, rbytes);
	}
})
#endif
#endif

/* ............................
 *
 * STOREs with implicit counter
 *
 * ............................
 */

EXTERN inline void __c_store(___V_GPTR gPtr, char val)
INLINE({ __c_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

EXTERN inline void __sh_store(___V_GPTR gPtr, short val)
INLINE({ __sh_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

EXTERN inline void __i_store(___V_GPTR gPtr, int val)
INLINE({
	__i_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

EXTERN inline void __f_store(___V_GPTR gPtr, float val)
INLINE({ __f_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

EXTERN inline void __d_store(___V_GPTR gPtr, double val)
INLINE({ __d_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

#ifdef __GNUC__
EXTERN inline void __ll_store(___V_GPTR gPtr, long long val)
INLINE({ __ll_store_ctr (gPtr, val, &recvBytes, &sentBytes); })
#endif

EXTERN inline void __struct_i_store(___V_GPTR gPtr, ___Struct_i val)
INLINE({ __i_store_ctr (gPtr, val.data, &recvBytes, &sentBytes); })

#ifdef __GNUC__
EXTERN inline void __struct_d_store(___V_GPTR gPtr, ___Struct_d val)
INLINE({ __ll_store_ctr (gPtr, val.data, &recvBytes, &sentBytes); })
#endif
	
extern void __store(___V_GPTR gPtr, void *lPtr, int len);

#endif __STORE_H
