/*                                                                      tab:8
 * malloc.c - management for global and local heaps.
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:              Steve Luna
 * Version:             1
 * Creation Date:       Tue Nov 30 22:41:55 1993
 * Filename:            malloc.c
 * History:
 *	LTL	1	Mon Jul 24 18:33:17 1995
 *		Wired down allocated memory for PAM
 *	ssl	1	Thu Apr 28 21:46:21 1994
 *		added dword align check in all_spread_malloc
 *	ssl	1	Sun Mar 20 20:25:50 1994
 *		modified to make initial sbrk dword aligned
 */

#include <sys/types.h>
#include <stdio.h>
#include <split-c/split-c.h>
#define SC_MALLOC_INTERNAL_DEFINE
#include "malloc_int.h"

extern caddr_t sbrk (int);

/*
 * Strategy: this is a binned memory allocator. There are separate bin tables
 * for global and local heaps. A bin table keeps a separate
 * free list for each bin size. Bin sizes are multiples of the bin granularity.
 * When a bin is empty, a fixed-size block (roughly a VM page) is allocated
 * and fragmented into the bin size. The per-bin free list is organized in
 * two levels: it is a list of blocks allocated for that bin size, and each
 * block holds a list of free fragments. Fragments larger than the largest
 * bin size are stored on a linked list. Elements on this list are allocated
 * using a best fit strategy. 
 */

sc_malloc_block __gsc_malloc_bin[SC_MALLOC_BINS];
sc_malloc_block __lsc_malloc_bin[SC_MALLOC_BINS];

/* chunks bigger that largest bin size are put on linked list */
sc_malloc_frag __gbig_queue;
sc_malloc_frag __lbig_queue;

static char *__top_heap;
static int init = 0;

/*
 * sc_init_malloc - init binned memory allocator
 */
static void sc_init_malloc (void)
{
	int b;
	sc_malloc_block *head;
	
	__top_heap = sbrk (0);
	/* must dword align */
	if ((int)__top_heap & 0x7)
	{
		sbrk(0x8 - (int)__top_heap & 0x7);
		__top_heap += 0x8 - (int)__top_heap & 0x7;
		if ((int)__top_heap & 0x7)
			printf("FATAL ERROR: Could not dword align heap\n");
	}
	__lbig_queue.chain.next = &__lbig_queue;
	__gbig_queue.chain.next = &__gbig_queue;
	
	for (b = 0; b < SC_MALLOC_BINS; b++)
	{
		head = &__gsc_malloc_bin[b];
		head->next = head;	/* note: head->next->free == 0 */
		head->head = head;
		head->free = 0;
		
		head = &__lsc_malloc_bin[b];
		head->next = head;	/* note: head->next->free == 0 */
		head->head = head;
		head->free = 0;
		
	}
}

/* insert a big fragment into big_queue */
static void enqueue_big (sc_malloc_frag * frag, unsigned long size, int global_flag)
{
	sc_malloc_frag *frag_p;
	frag_p = WHICH_QUEUE (global_flag);
	
	while (frag_p->chain.next != frag_p && frag_p->chain.next->size < size)
		frag_p = frag_p->chain.next;
	
	/* if frag should go at end of list */
	if (frag_p->chain.next == frag_p)
		frag->chain.next = frag;
	else
		frag->chain.next = frag_p->chain.next;
	frag_p->chain.next = frag;
	frag->size = size;
	frag->key = MAGIC_KEY;
}

/* get the first fragment from big queue that is big enough */
static sc_malloc_frag *dequeue_big (unsigned long size, int global_flag)
{
	sc_malloc_frag *frag_p, *frag;
	frag_p = WHICH_QUEUE (global_flag);
	
	while (frag_p->chain.next != frag_p && frag_p->chain.next->size < size)
		frag_p = frag_p->chain.next;
	
	/* none of the frags where big enough */
	if (frag_p->chain.next == frag_p)
		return NULL;
	
	frag = frag_p->chain.next;
	/* if last element will be taken out */
	if (frag_p->chain.next == frag_p->chain.next->chain.next)
		frag_p->chain.next = frag_p;
	else
		frag_p->chain.next = frag_p->chain.next->chain.next;
	frag->chain.block = 0;
	return frag;
}

/* make a block out of the bytes and attatch it to bin */
static void add_block (char *cp, unsigned long size)
{
	int bin, stride;
	unsigned long efrag;
	sc_malloc_frag *frag;
	sc_malloc_block *block;

	if (size > sizeof(sc_malloc_block))
		bin = (size - sizeof (sc_malloc_block)) /
			(SC_MALLOC_GRAN * SC_MALLOC_MINFRAGS);
	else
		bin = 0;

	/* ahh hell, it's so small - just throw it away */
	if (!bin)
		return;

	block = (sc_malloc_block *) cp;
	block->head = WHICH_BIN (0, bin);
	block->next = block->head->next;
	block->head->next = block;
	block->free = frag = (sc_malloc_frag *) (block + 1);

	efrag = (unsigned long) block + size - bin * SC_MALLOC_GRAN * 2;
	stride = bin * (SC_MALLOC_GRAN / sizeof (sc_malloc_frag));
	while ((unsigned long) frag <= efrag)
    {
		frag->chain.next = frag + stride;
		/* size of fragments in bins are only used for testing */
		frag->size = bin*SC_MALLOC_GRAN - sizeof(sc_malloc_frag);
		frag->key = MAGIC_KEY;
		frag += stride;
    }
	frag->chain.next = 0;
	frag->size = bin*SC_MALLOC_GRAN - sizeof(sc_malloc_frag);
	frag->key = MAGIC_KEY;
}

/* sbrk a chunk. if global, force the top of all heaps to be the same, then
   sbrk */
static char *chunk (unsigned long size, int global_flag)
{
	char *new_chunk, *max_heap;
	unsigned long frag_space;
	
	/* dword align allocations */
	size = ((size + 7) >> 3) << 3;
	if (global_flag)
    {
		/* bring all heaps up to same level */
		max_heap = (char *) all_reduce_to_all_umax ((unsigned)__top_heap);
		frag_space = (((unsigned)max_heap - (unsigned)__top_heap+7)>>3)<<3;
		if (frag_space > 0)
		{
			new_chunk = __top_heap;
			__top_heap += frag_space;
			if ((int) sbrk (frag_space) == -1)
			{
				fprintf (stderr, "FATAL error: Could not allocate sufficient "
						 "memory\n");
				exit (1);
			}
			/* give fragmented space to local heap */
			if (frag_space > SC_MAX_BIN_SIZE * SC_MALLOC_MINFRAGS)
				enqueue_big ((sc_malloc_frag *) new_chunk,
							 frag_space - sizeof (sc_malloc_frag), 0);
			else
				add_block (new_chunk, frag_space);
		}
    }
	new_chunk = __top_heap;
	__top_heap += size;
	if ((int) sbrk (size) == -1)
    {
		fprintf(stderr, "FATAL error: Could not allocate sufficient memory\n");
		exit (1);
    }
	return new_chunk;
}

/*
 * alloc_block - allocate block of memory and fragment
 */
static sc_malloc_block *alloc_block (unsigned long bin, int global_flag)
{
	sc_malloc_block *block;
	sc_malloc_frag *frag;
	unsigned long efrag;
	int size, stride;
	
	size = SC_MALLOC_MINFRAGS * bin * SC_MALLOC_GRAN
		+ sizeof (sc_malloc_block);
	if (size < SC_MALLOC_MINBLK)
		size = SC_MALLOC_MINBLK;
	block = (sc_malloc_block *) chunk (size, global_flag);
	if (!block)
		return 0;			/* PANIC!!! */
	
	block->next = 0;
	block->head = WHICH_BIN (global_flag, bin);
	
	block->free = frag = (sc_malloc_frag *) (block + 1);
	efrag = (unsigned long) block + size - bin * SC_MALLOC_GRAN * 2;
	stride = bin * (SC_MALLOC_GRAN / sizeof (sc_malloc_frag));
	while ((unsigned long) frag <= efrag)
    {
		frag->chain.next = frag + stride;
		/* size of fragments in bins are only used for testing */
		frag->size = bin * SC_MALLOC_GRAN - sizeof (sc_malloc_frag);
		frag->key = MAGIC_KEY;
		frag += stride;
    }
	frag->chain.next = 0;
	frag->size = bin * SC_MALLOC_GRAN - sizeof (sc_malloc_frag);
	frag->key = MAGIC_KEY;
	
	return block;
}

/*
 * sc_malloc - allocate memory
 */
static char *sc_malloc (unsigned long size, int global_flag)
{
	unsigned long bin;
	sc_malloc_block *head, *block;
	sc_malloc_frag *result;
	
	bin = (size + sizeof (sc_malloc_frag) + SC_MALLOC_GRAN - 1)
		>> SC_MALLOC_SHFT;
	if (bin >= SC_MALLOC_BINS)
    {
		result = dequeue_big (size, global_flag);
		if (!result)
		{
			result = (sc_malloc_frag *) chunk (size + sizeof (sc_malloc_frag),
											   global_flag);
			result->chain.block = 0;
			result->size = size;
			result->key = MAGIC_KEY;
		}
		return (void *) (result + 1);
    }
	
	/* give it a shot */
	head = WHICH_BIN (global_flag, bin);
	block = head->next;
	result = block->free;
	
	/* gotcha? */
	if (result)
    {
		block->free = result->chain.next;
		if (!block->free)
			head->next = block->next;	/* block all used up */
		result->chain.block = block;
		return (void *) (result + 1);
    }
	
	/* need new frag */
	if (size == 0)
		return 0;
	block = alloc_block (bin, global_flag);
	block->next = head;		/* alone in the block list */
	head->next = block;
	/* gotcha this time */
	result = block->free;
	block->free = result->chain.next;
	result->chain.block = block;
	return (void *) (result + 1);
}

/*
 * sc_free - free memory
 */
static void sc_free (char *addr, int global_flag)
{
	sc_malloc_frag *frag;
	sc_malloc_block *block, *head;

	if (!addr)
		return;
	
	frag = ((sc_malloc_frag *) addr) - 1;
	block = frag->chain.block;

	/* if it's not a valid frag */
	if (frag->key != MAGIC_KEY)
		return;
	
	/* is a big chunk? */
	if (!block)
    {
		enqueue_big (frag, frag->size, global_flag);
		return;
    }
	/* need to enqueue block? */
	if (!block->free)
    {
		head = block->head;
		/* put it second */
		/*block->next = head->next->next; */
		/*head->next->next = block; */
		/* put it first so tests that malloc free stuff will work */
		block->next = head->next;
		head->next = block;
    }
	/* enqueue frag */
	frag->chain.next = block->free;
	block->free = frag;
}

___V_SPTR all_spread_malloc (unsigned num_elems, unsigned bytes_per_elem)
{
	int s;
	unsigned size;
	static char *cp;

	if (!init)
	{
		sc_init_malloc();
		init++;
	}
	size = ((num_elems+PROCS-1)/PROCS) * bytes_per_elem;
	cp = sc_malloc (size, 1);
	/* do reduction after sc_malloc to ensure that any sbrks have been done */
	s = all_reduce_to_all_umax(size);
	if (size != s)
		printf("ERROR: all_spread_malloc called with different sizes\n");
	if ((int)cp & 0x7)
		printf("FATAL ERROR: all_spread_malloc returning address not dword"
			   "aligned\n");
#ifdef PAM
  if (mcmsg_wire(cp, size) != 0) {
		printf("Error: node %d cannot wire the memory region in all_spread_malloc() size %d\n", MYPROC, size);			
		exit(-1);
	}
#endif	
	return tospread(0, cp);
}

void all_spread_free (___V_SPTR sptr)
{
	char *c, *cp;

	cp = (char*)tolocal(sptr);
	c = (char*)all_reduce_to_all_umax((unsigned)cp);
	if (cp != c)
		printf("ERROR: all_spread_free called with different addresses\n");
	sc_free (cp, 1);
}

void *malloc(unsigned long size)
{
	char *cp;
	if (!init)
	{
		sc_init_malloc();
		init++;
	}
	cp = sc_malloc((unsigned long)size, 0);
	if ((int)cp <= 0)
		printf("FATAL ERROR: malloc fails\n");	
	if ((int)cp & 0x7)
		printf("FATAL ERROR: malloc returning address not dword aligned\n");
#ifdef PAM
  if (mcmsg_wire(cp, size) != 0) {
		printf("Error: node %d cannot wire the memory region in malloc() size %d\n", MYPROC, size);
		exit(-1);
	}
#endif		
	return (void*)cp;
}

void *malloc_nowire(unsigned long size)
{
	char *cp;
	if (!init)
	{
		sc_init_malloc();
		init++;
	}
	cp = sc_malloc((unsigned long)size, 0);
	if ((int)cp <= 0)
		printf("FATAL ERROR: malloc fails\n");	
	if ((int)cp & 0x7)
		printf("FATAL ERROR: malloc returning address not dword aligned\n");
	return (void*)cp;
}

void free(void *cp)
{
	sc_free((char*)cp, 0);
}

/* to appease ld */
char *realloc(char *ptr, unsigned size)
{
	printf("FATAL ERROR: realloc called\n");
}

