/*                                                                      tab:8
 *
 * atomic.sc - Atomic operations.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:             Andrea Dusseau
 * Version:            8         
 * Creation Date:      Thu Oct 22 15:49:07 PDT 1992
 * Filename:           atomic.sc
 * History:				
 *	ssl	1	Wed Jan 19 11:44:07 1994
 *		removed rpc
 *	ssl	8	Tue Nov 16 12:09:22 1993
 *		Replaced global tags with macros.
 *	ssl	7	Thu Oct 28 22:02:45 1993
 *		Include am.h instead of cm5.h
 *	LTL	6	Mon Aug 16 15:17:42 1993
 *		Fixed it up for version 112 to support CMMD
 *	SL	5	Wed Jan 20 11:10:14 1993
 *		        Added atomic and rpc long routines
 *	SL	4	Tue Jan 19 14:01:21 1993
 *		        Added atomic synchronization and rpc routines
 *      ACD     3       Thu Jan 14 15:42:03 PST 1993
 *              Removed "Static" qualifier from atomic_return_x_rh definitions
 *      ACD     2       Tue Jan  5 14:50:58 PST 1993
 *              Moved definitions for atomic, atomic_x from atomic_inline.h
 */

#include <stdio.h>
/*#include "am.h"*/
#include <split-c/split-c.h>
#define ATOMIC_INTERNAL_DEFINE
#include <split-c/atomic.h>


#define N_OUTSTANDING 2 /* Maximum number of long calls outstanding from
						   one processor to another */

typedef struct {
	unsigned id;
	int args[3];
} long_receive_t;

static int send_num = 0;
static long_receive_t rcv[PROCSMAX][N_OUTSTANDING];
	
static Counter atomic_sent = 0;
static atomic_t return_val;

/* ..........................................
 *
 * Initialization routine for long operations
 *
 * ..........................................
 */

void atomic_init()
{
	int i, j;

	for (i = 0; i < PROCS; i++)
		for (j = 0; j < N_OUTSTANDING; j++)	
			rcv[i][j].id = 0;
}

/* ...........................................
 *
 * Handlers for synchronized atomic operations
 *
 * ...........................................
 */

static void atomic_handler(int src_node, void (*fun)(), void *lptr, int arg2,
													 int arg3)
{
	(*fun)(lptr, arg2, arg3);
	atomic_sent--;
}

static void atomic_long_one_h(int src_node, unsigned id,
															void (*fun)(), void *lptr, int arg1)
{
	int from_proc = ((id>>20)&0x7FF);
	int msg_id = (0x80000000 | (id&0xFFFFF));
	int i;
	
	for (i = 0; i < N_OUTSTANDING; i++)
		if (rcv[from_proc][i].id == msg_id)
			break;
	if (i == N_OUTSTANDING) {
		for (i = 0; i < N_OUTSTANDING; i++)
			if (!(rcv[from_proc][i].id&0x80000000))
				break;
		if (i == N_OUTSTANDING) {
			printf ("Internal error--atomic long overflow\n");
			exit (-3);
		}
		rcv[from_proc][i].id = msg_id;
		rcv[from_proc][i].args[0] = (int)fun;
		rcv[from_proc][i].args[1] = (int)lptr;
		rcv[from_proc][i].args[2] = arg1;
		return;
	}
	(*fun)(lptr, arg1, rcv[from_proc][i].args[0], rcv[from_proc][i].args[1],
		   rcv[from_proc][i].args[2]);
	rcv[from_proc][i].id = 0;
	atomic_sent--;
}

static void atomic_long_two_h(int src_node, unsigned id, int arg2,
															int arg3, int arg4)
{
	int from_proc = ((id>>20)&0x7FF);
	int msg_id = (0x80000000 | (id&0xFFFFF));
	int i;
	
	for (i = 0; i < N_OUTSTANDING; i++)
		if (rcv[from_proc][i].id == msg_id)
			break;
	if (i == N_OUTSTANDING) {
		for (i = 0; i < N_OUTSTANDING; i++)
			if (!(rcv[from_proc][i].id&0x80000000))
				break;
		if (i == N_OUTSTANDING) {
			printf ("Internal error--atomic long overflow\n");
			exit (-3);
		}
		rcv[from_proc][i].id = msg_id;
		rcv[from_proc][i].args[0] = arg2;
		rcv[from_proc][i].args[1] = arg3;
		rcv[from_proc][i].args[2] = arg4;
		return;
	}
	(*((void (*)())rcv[from_proc][i].args[0]))(rcv[from_proc][i].args[1],
											   rcv[from_proc][i].args[2],
											   arg2, arg3, arg4);
	rcv[from_proc][i].id = 0;
	atomic_sent--;
}

/* .............................................
 *
 * Building blocks for general atomic operations
 *
 * .............................................
 */

#ifdef PAM
static void atomic_no_sync_handler(int src_node, void (*fun)(),
																	 void *lptr, int arg2, int arg3, int arg4)
{
	(*fun)(lptr, arg2, arg3, arg4);
}

void atomic_no_sync(void (*fun)(), ___V_GPTR gptr, int arg2, int arg3,
					int arg4)
INLINE({
	am_request_5(toproc(gptr), atomic_no_sync_handler,
							 fun, tolocal(gptr), arg2, arg3, arg4);
})

#else
void atomic_no_sync(void (*fun)(), ___V_GPTR gptr, int arg2, int arg3,
					int arg4)
INLINE({
	am_request_4(toproc(gptr), fun, tolocal(gptr), arg2, arg3, arg4);
})
#endif

void atomic(void (*fun)(), ___V_GPTR gptr, int arg3, int arg4)
INLINE({
	am_request_4(toproc(gptr), atomic_handler, fun, tolocal(gptr), arg3, arg4);
	atomic_sent++;
})

void atomic_long(void (*fun)(), ___V_GPTR gptr, int arg1, int arg2,
				 int arg3, int arg4)
INLINE({
	unsigned id = (0x80000000 | (MYPROC<<20) | send_num);

	am_request_4(toproc(gptr), atomic_long_one_h, id, fun, tolocal(gptr), arg1);
	am_request_4(toproc(gptr), atomic_long_two_h, id, arg2, arg3, arg4);
	atomic_sent++;
	send_num = (send_num + 1)&0xFFFFF;
})

int atomic_i(void (*fun)(), ___V_GPTR gptr, int arg3, int arg4)
INLINE({
	return_val.flag = 0;

	am_request_3(toproc(gptr), fun, tolocal(gptr), arg3, arg4);
	am_wait(&return_val.flag, 1);
	return return_val.value.v_int;
})

#if 0
float atomic_f(void (*fun)(), ___V_GPTR gptr, int arg3, int arg4)
INLINE({
	return_val.flag = 0;

	am_request_3(toproc(gptr), fun, tolocal(gptr), arg3, arg4);
	am_wait(&return_val.flag, 1);
	return return_val.value.v_float;
})
#endif

double atomic_d(void (*fun)(), ___V_GPTR gptr, int arg3, int arg4)
INLINE({
	return_val.flag = 0;

	am_request_3(toproc(gptr), fun, tolocal(gptr), arg3, arg4);
	am_wait(&return_val.flag, 1);
	return return_val.value.v_double;
})

/* ...........................................
 *
 * Handlers for replies from atomic operations
 *
 * ...........................................
 */

void atomic_return_i_rh(int src_node, int val)
{
	return_val.value.v_int = val;
	return_val.flag = 1;
}

void atomic_return_d_rh(int src_node, double val)
{
	return_val.value.v_double = val;
	return_val.flag = 1;
}

/* .................................................
 *
 * Handlers for canned atomic operations on integers
 *
 * .................................................
 */

void cmp_and_swap_atomic(int proc, short *loc, int test, int new)
{
	if(*loc == test) {
		*loc = new;
		atomic_return_i(1);
	}
	atomic_return_i(0);
}

void test_and_set_atomic(int proc, int *loc)
{
	int tmp = *loc;
	*loc = 1;
	atomic_return_i(tmp);
}

void exchange_atomic(int proc, int *loc, int new_val)
{
	int tmp = *loc;
	*loc = new_val;
	atomic_return_i(tmp);
}

void fetch_and_add_atomic(int proc, int *loc, int new_val)
{
	int tmp = *loc;
	*loc += new_val;
	atomic_return_i(tmp);
}

/* .........................................
 *
 * Atomic and rpc synchronization primitives
 *
 * .........................................
 */

/* Wait until all calls to atomic from all processors have completed */
void all_atomic_sync()
{
	int res;
	extern int ___i_reduce(int diff);

	am_barrier(0);
	do {
		res = ___i_reduce(atomic_sent);
	} while(res != 0);

	atomic_sent = 0;
}

