/*									tab:8
 *
 * sc-transform.c - tree generation codes for split-c
 *
 * Copyright (c) 1992 by Seth Copen Goldstein and 
 * Regents of the University of California
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Copyright holder(s) make no
 * representation about the suitability of this software for
 * any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * Author: 			Seth Copen Goldstein
 * Version:			143
 * Creation Date:	Tue Dec 15 13:51:22 1992
 * Filename:		sc-transform.c
 * History:
 *	AK	105	Tue Jul  6 14:33:56 1993
 *		Moved a lot of code from c-parse.y
 *	AK	104	Tue Jul  6 14:33:33 1993
 *		Changed the way spread arrays are handled.
 *	AK	93	Mon Jun  7 18:17:34 1993
 *	        Removed exact_log2 decln from set_pe_knowledge.
 *	AK	90	Mon Jun  7 16:02:11 1993
 *		Changed the type of enum tree_code in build_modify_expr_with_opt
 *	AK	83	Sun Jan 24 11:58:56 1993
 *		Added routines to transform reads and postinc/postdec opns.
 *	AK	70	Thu Jan 21 14:24:50 1993
 *		Added the transformation for rvalue writes.
 *	AK	54	Tue Jan 19 13:51:16 1993
 *		Changed sizeof_int_or_real routine to take care of unions
 * 	TvE	43	Tue Jan 19 00:39:14 PST 1993
 *		Removed knowledge of power of two-ness.
 * 	TvE	38	Thu Dec 24 05:33:54 PST 1992
 *
 *      SCG	1	Tue Dec 15 13:51:22 1992
 *		Obtained from c-parse.y
 */
static char _version_string_[] = "\nVersion:105:split-c.c\0\n";

# include <stdio.h>
# include <errno.h>
# include <setjmp.h>

# include "config.h"
# include "tree.h"
# include "input.h"
# include "c-lex.h"
# include "c-tree.h"
# include "flags.h"

# ifdef MULTIBYTE_CHARS
#  include <stdlib.h>
#  include <locale.h>
# endif
#include "split-c.h"

enum type_len {
    CHAR_LEN, SHORT_LEN, INT_LEN, LONG_LEN, FLOAT_LEN, 
    DOUBLE_LEN, STRUCT_I_LEN, STRUCT_D_LEN, STRUCT_LEN
};

/***************************************************************************
 * builds nodes that refers to the number of processors in the system.  This
 * will be specialized to handle cases when we know the number of pes at
 * compile time, when we only know they are a power of 2, and when we know
 * nothing.
 ****************************************************************************/
/* this code assumes that the following are defined before any special split-c
 * code is compiled:
 *		const int ___PROCS;		== (partition size)
 *		const int ___PROCSLOG;		== exact_log2(___PROCS)
 */

/* The number of processors for which the compilation is taking place */

static int numPes =    -1;         /* -1 if we don't know */
static int logP  = 	0;	   /* 1 if compiling for 2^N procs, 0 if not. */

tree build_numPEs(void)
{
	static tree id = NULL;
	if (numPes != -1) return build_int_2 (numPes, 0);
	if (id == NULL) 
		id = lookup_splitc("___PROCS");
	return id;
}

/* call this routine before generating code so that we make the right choice
 * about what kind of node to build for things like the number of processors,
 * etc.
 */
void set_pe_knowledge (char *buffer)
{
	int n;

	if (strncmp(buffer, "PROCSLOG=", 9) == 0) {
	    /* number of processors is fixed power-of-two */
	    sscanf (buffer+9, "%d", &n);
	    if (n < 0 || n > 30) {
	        error ("Illegal value for -fPROCSLOG: %d", n);
	        return;
	    }
	    printf ("Compiling for 2^%d processors\n", n);
	    numPes = 1 << n;
	    logP = 1;
	} else if (strncmp(buffer, "PROCSLOG", 8) == 0) {
	    /* number of processors is unknown power-of-two */
	    printf ("Compiling for 2^N processors\n");
	    numPes = -1;
	    logP = 1;
	} else if (strncmp(buffer, "PROCS=", 6) == 0) {
	    /* number of processors is fixed non-power-of-two */
	    sscanf (buffer+6, "%d", &n);
	    if (n <= 0 || n > (1<<30)) {
	        error ("Illegal value for -fPROCS: %d", n);
	        return;
	    }
	    numPes = n;
	    printf ("Compiling for %d processors\n", numPes);
	    if (exact_log2(numPes) >= 0) {
	        warning ("For optimal performance, use -fPROCSLOG=%d",					exact_log2(numPes));
	    }
	    logP = 0;
	} else if (strncmp(buffer, "PROCS", 5) == 0) {
	    /* number of processors is unknown non-power-of-two */
	    printf ("Compiling for N processors\n");
	    numPes = -1;
	    logP = 0;
	} else {
	    error ("Unrecognized -fPROCSLOG or -fPROCS flag: `%s'", buffer);
	}
}

extern tree build_function_call1 ();

/* These are aux. functions intoduced to handle global stuff
   - should be moved to c-typeck.c				*/

/* Make an expression to refer to the COMPONENT field of a pointer to a
   structure or union value DATUM.  COMPONENT is an IDENTIFIER_NODE.  */

tree
build_global_component_ref (datum, component, comp_ptr_type)
     tree datum, component;
     tree *comp_ptr_type;
{

  /* datum is a pointer - we are actually doing p->x stuff here.
     TREE_TYPE (TREE_TYPE (datum)) is the actual record type  */
  register tree type = TREE_TYPE (TREE_TYPE (datum));
  register enum tree_code code = TREE_CODE (type);
  register tree field = NULL;
  register tree ref;
  register int globalp = TYPE_GLOBAL (TREE_TYPE (datum));
  register int splitp = TYPE_SPLIT (TREE_TYPE (datum));
  register int distp = (globalp ? 2 : (splitp ? 3 : 0));
  register tree temp_type;

  /* See if there is a field or component with name COMPONENT.  */

  if (code == RECORD_TYPE || code == UNION_TYPE)
    {
      if (TYPE_SIZE (type) == 0)
	{
	  incomplete_type_error (0, type);
	  return error_mark_node;
	}

      /* Look up component name in the structure type definition.

	 If TYPE_LANG_SPECIFIC is set, then it is a sorted array of pointers
	 to the field elements.  Use a binary search on this array to quickly
	 find the element.  Otherwise, do a linear search.  TYPE_LANG_SPECIFIC
	 will always be set for structures which have many elements.  */

      if (TYPE_LANG_SPECIFIC (type))
	{
	  int bot, top, half;
	  tree *field_array = &TYPE_LANG_SPECIFIC (type)->elts[0];

	  field = TYPE_FIELDS (type);
	  bot = 0;
	  top = TYPE_LANG_SPECIFIC (type)->len;
	  while (top - bot > 1)
	    {
	      int cmp;

	      half = (top - bot + 1) >> 1;
	      field = field_array[bot+half];
	      cmp = (long)DECL_NAME (field) - (long)component;
	      if (cmp == 0)
		break;
	      if (cmp < 0)
		bot += half;
	      else
		top = bot + half;
	    }

	  if (DECL_NAME (field_array[bot]) == component)
	    field = field_array[bot];
	  else if (DECL_NAME (field) != component)
	    field = 0;
	}
      else
	{
	  for (field = TYPE_FIELDS (type); field; field = TREE_CHAIN (field))
	    {
	      if (DECL_NAME (field) == component)
		break;
	    }
	}

      if (!field)
	{
	  error (code == RECORD_TYPE
		 ? "structure has no member named `%s'"
		 : "union has no member named `%s'",
		 IDENTIFIER_POINTER (component));
	  return error_mark_node;
	}
      if (TREE_TYPE (field) == error_mark_node)
	return error_mark_node;

/* Change for the structure version - AK */
      ref = build_address_increments (datum, 
				      size_binop (CEIL_DIV_EXPR, DECL_FIELD_BITPOS (field),
						  size_int (TYPE_PRECISION (char_type_node))));

      *comp_ptr_type = build_pointer_type (TREE_TYPE (field));
      temp_type = c_build_type_variant (build_pointer_type (TREE_TYPE (field)), 
					0, 0, distp);

      TREE_TYPE (ref) = temp_type;

      if (TREE_READONLY (datum) || TREE_READONLY (field))
	TREE_READONLY (ref) = 1;
      if (TREE_THIS_VOLATILE (datum) || TREE_THIS_VOLATILE (field))
	TREE_THIS_VOLATILE (ref) = 1;

      return ref;
    }
  else if (code != ERROR_MARK)
    error ("request for member `%s' in something not a structure or union",
	    IDENTIFIER_POINTER (component));

  return error_mark_node;
}

/* This is for bulk transfer - it copies a huge chunk from either global
   to local or the other way around. Not exactly clear whats to be done */
tree
build_xfer (globPtr, start, end, stride)
     tree globPtr, start, end, stride;
{

}

/* This just converts the get on the lhs to a put and attaches
   the value to be put as an extra parameter at the front	
   - note dest is actually a function call and has params etc */
tree
build_global_put (dest, source, nonBlock)
     tree dest, source;
     int  nonBlock;
{

    tree param_list;
    tree value;
    tree temp_node;
    tree funName;
    int  size;
    tree gPtr;

    dest = sc_default_conversion (dest);
    source = sc_default_conversion (source);

    gPtr = dest;
    temp_node = TREE_TYPE (TREE_TYPE (gPtr));
    size = sizeof_int_or_real (temp_node);

    /* Convert new value to destination type.  */
    source = convert_for_assignment (temp_node, source, "assignment",
				     NULL_TREE, 0);
    if (TREE_CODE (source) == ERROR_MARK)
	return error_mark_node;

    switch (size) {

      case CHAR_LEN:
	        if (!nonBlock) funName = lookup_splitc ("__c_write");
		else if (nonBlock == 1) funName = lookup_splitc ("__c_put");
		else funName = lookup_splitc ("__c_store");

	        break;
		
      case SHORT_LEN:
	        if (!nonBlock) funName = lookup_splitc ("__sh_write");
		else if (nonBlock == 1) funName = lookup_splitc ("__sh_put");
		else funName = lookup_splitc ("__sh_store");

	        break;

      case INT_LEN:
	        if (!nonBlock) funName = lookup_splitc ("__i_write");
		else if (nonBlock == 1) funName = lookup_splitc ("__i_put");
		else funName = lookup_splitc ("__i_store");

	        break;

      case FLOAT_LEN:
                if (!nonBlock) funName = lookup_splitc ("__f_write");
		else if (nonBlock == 1) funName = lookup_splitc ("__f_put");
		else funName = lookup_splitc ("__f_store");

	        break;


      case STRUCT_I_LEN:
	        if (!nonBlock) funName = lookup_splitc ("__struct_i_write"); 
	        else if (nonBlock == 1) funName = lookup_splitc ("__struct_i_put");
	        else funName = lookup_splitc ("__struct_i_store");

	        break;

      case STRUCT_D_LEN:
		if (!nonBlock) funName = lookup_splitc ("__struct_d_write"); 
	        else if (nonBlock == 1) funName = lookup_splitc ("__struct_d_put");
		else funName = lookup_splitc ("__struct_d_store");

	        break;

      case LONG_LEN:
                if (!nonBlock) funName = lookup_splitc ("__ll_write");
		else if (nonBlock == 1) funName = lookup_splitc ("__ll_put");
		else funName = lookup_splitc ("__ll_store");

	        break;

      case DOUBLE_LEN:
	        if (!nonBlock) funName = lookup_splitc ("__d_write");
		else if (nonBlock == 1) funName = lookup_splitc ("__d_put");
		else funName = lookup_splitc ("__d_store");

	        break;

      /* for non-4 or 8 byte values pass a pointer to the source - would
	 not work for non-lvalue sources - has to be modified later */
      case STRUCT_LEN:	
                if (!nonBlock) funName = lookup_splitc ("__write"); 
	        else if (nonBlock == 1) funName = lookup_splitc ("__put");
		else funName = lookup_splitc ("__store");

                /* Check whether the source is an l-value or not 
		   If it is not an lvalue then pass it to a blocking send fun */
		if (!lvalue_p (source)) {
		    if (nonBlock) {
			error ("L-value expected for non-blocking operation");
			return error_mark_node;
		    }

#if 0
		    funName = lookup_splitc ("__write_rvalue");
		    error ("Not currently implemented - rewrite the statement to supply an lvalue in the rhs");

		    param_list = chainon (chainon (build_tree_list (NULL_TREE, gPtr),
						   build_tree_list (NULL_TREE, 
								    c_sizeof (temp_node))),
					  build_tree_list (NULL_TREE, source));

		    return build_function_call1 (funName, param_list);
#endif
		    return transform_rvalue_writes (gPtr, source);

		}
		else {
		    param_list 
			= chainon (chainon (build_tree_list(NULL_TREE, gPtr),
					    build_tree_list(NULL_TREE,
							    build_unary_op (ADDR_EXPR, source, 0))),
				   build_tree_list (NULL_TREE, c_sizeof (temp_node)));
		    return build_function_call1 (funName, param_list);
		}		    
                break;

    }
    /* This is only for the basic types */
    return build_function_call1 (funName,
				chainon (build_tree_list (NULL_TREE, gPtr),
					 build_tree_list (NULL_TREE, source)));
}

/* Take the global pointer - the first argument and make it 
   point to the field of the structure - latest additions  */
tree
modify_get_call (funcal, comp)
     tree funcal, comp;
{

    tree new_ref;
    tree pid_node;
    tree value;
    tree funName;
    tree temp_node, temp_type;
    int  size;

    funcal = sc_default_conversion (funcal);
    /* the function call would have returned a pointer to struct - strip of
       the pointer - by doing TREE_OPERAND and then do TREE_OPERAND again
       to get the pointer originally passed to get */

    if (TREE_CODE (funcal) == INDIRECT_REF)
	funcal = TREE_OPERAND (funcal, 0);

    /* global pointer is the first argument to get */
    new_ref = build_global_component_ref (TREE_VALUE (TREE_OPERAND (funcal,1)), comp, &temp_type);

    if (new_ref == error_mark_node) return new_ref;

    /* new_ref has pointer_type - we get the type of the record */
    temp_node = TREE_TYPE (TREE_TYPE (new_ref));

    /* If returning an array the value remains the same - just the type
       changes. The lang-flag-3 is also set and we follow the split-ref appr */
    if (TREE_CODE (temp_node) == ARRAY_TYPE) {
	value = new_ref;

#ifdef NOMORE
	if (TYPE_LANG_FLAG_3 (temp_node))
	    TREE_TYPE (value) = c_build_type_variant (build_pointer_type (TREE_TYPE (temp_node)),
						      0, 0, 0, 1);
	else
	    TREE_TYPE (value) = c_build_type_variant (build_pointer_type (TREE_TYPE (temp_node)),
						      0, 0, 1, 0);
#endif
	TREE_TYPE (value) = temp_node;

	if (!TYPE_LANG_FLAG_3 (temp_node))
	    TREE_LANG_FLAG_3 (value) = 1;

	return value;
    }

    /* make call to the appropriate function */
    size = sizeof_int_or_real (temp_node);
    switch (size) {

      case CHAR_LEN:	funName = lookup_splitc ("__c_read"); break;

      case SHORT_LEN:	funName = lookup_splitc ("__sh_read"); break;

      case INT_LEN: 	funName = lookup_splitc ("__i_read"); break;

      case LONG_LEN:	funName = lookup_splitc ("__ll_read"); break;

      case FLOAT_LEN:	funName = lookup_splitc ("__f_read"); break;

      case DOUBLE_LEN:	funName = lookup_splitc ("__d_read"); break;

      case STRUCT_I_LEN:funName = lookup_splitc ("__struct_i_read"); break;

      case STRUCT_D_LEN:funName = lookup_splitc ("__struct_d_read"); break;

      case STRUCT_LEN:	funName = lookup_splitc ("__read"); break;

    }

    /* set the type of the value returned by the 
       function to that of type of new_ref       */
    if (size == CHAR_LEN || size == SHORT_LEN || size == INT_LEN || 
	size == LONG_LEN || size == FLOAT_LEN || size == DOUBLE_LEN) {
	value = build_function_call1 (funName, build_tree_list (NULL_TREE, new_ref));
	TREE_TYPE (value) = temp_node;
    }
    else {
	if (size == STRUCT_LEN)
	    value = build_function_call1 (funName,
					  chainon (build_tree_list (NULL_TREE, new_ref),
						   build_tree_list (NULL_TREE, 
								    c_sizeof (temp_node))));
	else
	    value = build_function_call1 (funName, build_tree_list (NULL_TREE, new_ref));

	TREE_TYPE (value) = build_pointer_type (temp_node);
	value = build_indirect_ref (value, "unary *");
    }

    return value;

}

tree
build_global_get (glob_ptr)
     tree glob_ptr;
{

    tree value;
    tree funName;
    tree temp_node, newType;
    int  size;


    glob_ptr = sc_default_conversion (glob_ptr);

    temp_node = TREE_TYPE (TREE_TYPE (glob_ptr));

    /* If returning an array the value remains the same - just the type
       changes. The lang-flag-3 is also set and we follow the split-ref appr */
    if (TREE_CODE (temp_node) == ARRAY_TYPE) {

#ifdef NOMORE
	if (TYPE_LANG_FLAG_3 (temp_node))
	    newType = c_build_type_variant (build_pointer_type (TREE_TYPE (temp_node)),
						      0, 0, 0, 1);
	else
	    newType = c_build_type_variant (build_pointer_type (TREE_TYPE (temp_node)),
						      0, 0, 1, 0);
#endif
	newType = temp_node;
	value = build1 (NOP_EXPR, newType, glob_ptr);
	if (!TYPE_LANG_FLAG_3 (temp_node)) {
	    TREE_LANG_FLAG_3 (value) = 1;
	}
	return value;
    }

    /* make call to the appropriate function */
    size = sizeof_int_or_real (temp_node);

    switch (size) {

      case CHAR_LEN:	funName = lookup_splitc ("__c_read"); break;

      case SHORT_LEN:	funName = lookup_splitc ("__sh_read"); break;

      case INT_LEN: 	funName = lookup_splitc ("__i_read"); break;

      case LONG_LEN:	funName = lookup_splitc ("__ll_read"); break;

      case FLOAT_LEN:	funName = lookup_splitc ("__f_read"); break;

      case DOUBLE_LEN:	funName = lookup_splitc ("__d_read"); break;

      case STRUCT_I_LEN:	funName = lookup_splitc ("__struct_i_read"); break;

      case STRUCT_D_LEN:funName = lookup_splitc ("__struct_d_read"); break;

      case STRUCT_LEN:	funName = lookup_splitc ("__read"); break;

    }

    /* set the type of the value returned by the 
       function to that of type of new_ref       */
    if (size == CHAR_LEN || size == SHORT_LEN || size == INT_LEN || 
	size == LONG_LEN || size == FLOAT_LEN || size == DOUBLE_LEN) {
	value = build_function_call1 (funName, build_tree_list (NULL_TREE, glob_ptr));
	TREE_TYPE (value) = temp_node;
    }
    else {
	if (size == STRUCT_LEN)
	    value = build_function_call1 (funName,
					  chainon (build_tree_list (NULL_TREE, glob_ptr),
						   build_tree_list (NULL_TREE, 
								    c_sizeof (temp_node))));
	else
	    value = build_function_call1 (funName, build_tree_list (NULL_TREE, glob_ptr));
	
	TREE_TYPE (value) = build_pointer_type (temp_node);
	value = build_indirect_ref (value, "unary *");
	TREE_TYPE (value) = temp_node;
    }

    return value;

}

tree
build_global_component_get (glob_ptr, comp)
     tree glob_ptr, comp;
{

    return modify_get_call (build_global_get (glob_ptr), comp);

}

/* the only diff between this and get is that there is some pointer
   arithmetic due to the index - the code has obviously been block copied */
tree
build_global_array_ref (glob_ptr, index)
     tree glob_ptr, index;
{

    tree temp_node;
    tree size_node;

#if 0 /* COME BACK */
    if (TYPE_GLOBAL (TREE_TYPE (glob_ptr))) {
	size_node = parser_build_binary_op (MULT_EXPR,
					    index,
					    c_sizeof (TREE_TYPE (TREE_TYPE (glob_ptr))));
	temp_node = parser_build_binary_op (PLUS_EXPR, glob_ptr, size_node);
	TREE_TYPE (temp_node) = TREE_TYPE (glob_ptr);


	return build_global_get (temp_node);
    }

    glob_ptr = build_split_arith_op (glob_ptr, index);
#endif

    glob_ptr = sc_default_conversion (glob_ptr);

    glob_ptr = build_global_ptr_arith (PLUS_EXPR, glob_ptr, index);
    return build_global_get (glob_ptr);

}

/* handles casting local pointers to global pointers
   makes a call to the global_cast function which would
   set my pid value as the last 4 bytes			*/
tree
build_sc_cast (cast_type, castee)
     tree cast_type, castee;
{
    tree value;

    castee = sc_default_conversion (castee);
    if (!(TYPE_GLOBAL_PTR (cast_type)
	  || (TREE_TYPE (cast_type) && TYPE_GLOBAL_PTR (TREE_TYPE (castee)))))
	return build_c_cast (cast_type, castee);

    if (TYPE_GLOBAL_PTR (cast_type) && TYPE_GLOBAL_PTR (TREE_TYPE (castee)))
	return build1 (NOP_EXPR, cast_type, castee);
    else if (!TYPE_GLOBAL_PTR (TREE_TYPE (castee))) {

	if (TREE_CODE (TREE_TYPE (castee)) == ARRAY_TYPE)
	    castee = default_conversion (castee);

	value = build_function_call (lookup_splitc ("___global_cast"),
				     build_tree_list (NULL_TREE, castee));
	TREE_TYPE (value) = cast_type;
	return value;

    }

    /* This is for the global to local pointer conversion. */
    value = build_function_call1 (lookup_splitc ("tolocal"),
				  build_tree_list (NULL_TREE, castee));
    
    TREE_TYPE (value) = cast_type;
    return value;

/*    return build1 (CONVERT_EXPR, cast_type, value);    */

}

/* optimize for the case that you are returning a struct and copying
   it to a local struct - this would mean that you can pass local ptr to get */
tree
build_modify_expr_with_opt (dest, op, source, nonBlock)
     tree dest, source;
     enum tree_code op;
     int  nonBlock;
{

    tree funName;
    tree param_list;
    tree funcall;
    tree temp_node;
    tree size;

    dest = sc_default_conversion (dest);
    source = sc_default_conversion (source);

    /* COME BACK - make type checks */
    if (!nonBlock) {

	if (TREE_CODE (source) == INDIRECT_REF && TREE_OPERAND (source,0)) {
	    if (TREE_CODE (TREE_OPERAND (source,0)) == CALL_EXPR) {
		
		param_list = TREE_OPERAND (TREE_OPERAND (source, 0), 1);
		funcall = TREE_OPERAND (TREE_OPERAND (source, 0), 0);
		
		if (TREE_OPERAND (funcall, 0) == lookup_splitc ("__read"))
		    funName = lookup_splitc ("__read_local");
		else
		    return build_modify_expr (dest, op, source);

		{
		    /* The global pointer from which the get takes place */
		    tree gPtr;
		    tree lPtr;
		    
		    gPtr = TREE_VALUE (param_list);
		    lPtr = build_unary_op (ADDR_EXPR, dest, 0);
		    size = c_sizeof (TREE_TYPE (TREE_TYPE (lPtr)));
		    
		    return build_function_call1 (funName, 
						chainon (chainon (build_tree_list (NULL_TREE, lPtr),
								  build_tree_list (NULL_TREE, gPtr)),
							 build_tree_list (NULL_TREE, size)));
							 
		}
	    }
	}
	return build_modify_expr (dest, op, source);
    }

    if (TREE_CODE (source) == INDIRECT_REF && TREE_OPERAND (source,0)) {
	if (TREE_CODE (TREE_OPERAND (source,0)) == CALL_EXPR) {
	    
	    param_list = TREE_OPERAND (TREE_OPERAND (source, 0), 1);
	    funcall = TREE_OPERAND (TREE_OPERAND (source, 0), 0);
	    
	    if (TREE_OPERAND (funcall, 0) == lookup_splitc ("__read"))
		funName = lookup_splitc ("__get");
	    else if (TREE_OPERAND (funcall, 0) == lookup_splitc ("__struct_i_read"))
		funName = lookup_splitc ("__struct_i_get");
	    else if (TREE_OPERAND (funcall, 0) == lookup_splitc ("__struct_d_read"))
		funName = lookup_splitc ("__struct_d_get");
	    else
		return build_modify_expr (dest, op, source);

	    {
		/* The global pointer from which the get takes place */
		tree gPtr;
		tree lPtr;
		
		gPtr = TREE_VALUE (param_list);
		lPtr = build_unary_op (ADDR_EXPR, dest, 0);
		size = c_sizeof (TREE_TYPE (TREE_TYPE (lPtr)));
		
		return build_function_call1 (funName, 
					    chainon (chainon (build_tree_list (NULL_TREE, lPtr),
							      build_tree_list (NULL_TREE, gPtr)),
						     build_tree_list (NULL_TREE, size)));
	    }
	}
    }
    else if (TREE_CODE (source) == CALL_EXPR) {
	
	param_list = TREE_OPERAND (source, 1);
	funcall = TREE_OPERAND (source, 0);
	
	if (TREE_OPERAND (funcall, 0) == lookup_splitc ("__c_read"))
	    funName = lookup_splitc ("__c_get");
	else if (TREE_OPERAND (funcall, 0) == lookup_splitc ("__sh_read"))
	    funName = lookup_splitc ("__sh_get");
	else if (TREE_OPERAND (funcall, 0) == lookup_splitc ("__i_read"))
	    funName = lookup_splitc ("__i_get");
	else if (TREE_OPERAND (funcall, 0) == lookup_splitc ("__ll_read"))
	    funName = lookup_splitc ("__ll_get");
	else if (TREE_OPERAND (funcall, 0) == lookup_splitc ("__f_read"))
	    funName = lookup_splitc ("__f_get");
	else if (TREE_OPERAND (funcall, 0) == lookup_splitc ("__d_read"))
	    funName = lookup_splitc ("__d_get");
	else
	    return build_modify_expr (dest, op, source);

	{
	    /* The global pointer from which the get takes place */
	    tree gPtr;
	    tree lPtr;

	    gPtr = TREE_VALUE (param_list);
	    lPtr = build_unary_op (ADDR_EXPR, dest, 0);

	    return build_function_call1 (funName, chainon (build_tree_list (NULL_TREE, lPtr),
							  build_tree_list (NULL_TREE, gPtr)));

	}
	
    }
    
    return build_modify_expr (dest, op, source);

}

int
sizeof_int_or_real (type)
     tree type;
{

    if (TREE_CODE (type) == CHAR_TYPE || TREE_INT_CST_LOW (TYPE_SIZE (type)) == 8)
	return CHAR_LEN;

    /* Dont make a check for integer_type */
    if (TREE_CODE (type) == INTEGER_TYPE && TREE_INT_CST_LOW (TYPE_SIZE (type)) <= 16)
	return SHORT_LEN;

    if (TREE_CODE (type) == INTEGER_TYPE && TREE_INT_CST_LOW (TYPE_SIZE (type)) <= 32)
	return INT_LEN; /* single word packet required */

    if (TREE_CODE (type) == INTEGER_TYPE && TREE_INT_CST_LOW (TYPE_SIZE (type)) <= 64)
	return LONG_LEN; /* double word packet required */

    if (TREE_CODE (type) == REAL_TYPE && TREE_INT_CST_LOW (TYPE_SIZE (type)) <= 32)
	return FLOAT_LEN; /* single word packet required */

    if (TREE_CODE (type) == REAL_TYPE && TREE_INT_CST_LOW (TYPE_SIZE (type)) <= 64)
	return DOUBLE_LEN; /* double word packet required */

#if 0
    return STRUCT_LEN; /* An hack now to get things working */
#endif

    if (TREE_INT_CST_LOW (TYPE_SIZE (type)) <= 32)
	if (TREE_CODE (type) == RECORD_TYPE ||
	    TREE_CODE (type) == UNION_TYPE)
	    return STRUCT_I_LEN; /* single word packet required */
	else
	    return INT_LEN;

    if (TREE_INT_CST_LOW (TYPE_SIZE (type)) <= 64)
	if (TREE_CODE (type) == RECORD_TYPE ||
	    TREE_CODE (type) == UNION_TYPE)
	    return STRUCT_D_LEN; /* double word packet required */
	else
	    return LONG_LEN;

#if 0
    if (TREE_INT_CST_LOW (TYPE_SIZE (type)) <= 96)
	return 7; /* Might not require xfer if no-flag */
#endif

    return STRUCT_LEN; /* xfer required - to be optimized later */

}

/* Checks whether it is a global operation or not */
int
global_op (expr)
     tree expr;
{

    if (TREE_CODE (expr) == CALL_EXPR)
	if (  TREE_OPERAND (TREE_OPERAND (expr, 0),0) == 
	         lookup_splitc ("__c_read")
	    ||TREE_OPERAND (TREE_OPERAND (expr, 0),0) == 
	         lookup_splitc ("__sh_read")
	    ||TREE_OPERAND (TREE_OPERAND (expr, 0),0) == 
	         lookup_splitc ("__i_read")
	    ||TREE_OPERAND (TREE_OPERAND (expr, 0),0) == 
	         lookup_splitc ("__ll_read")
	    ||TREE_OPERAND (TREE_OPERAND (expr, 0),0) == 
	         lookup_splitc ("__f_read")
	    ||TREE_OPERAND (TREE_OPERAND (expr, 0),0) == 
	         lookup_splitc ("__d_read"))
	    return 1;
	else 
	    return 0;
    else
	if ((TREE_CODE (expr) == INDIRECT_REF) &&
	    (TREE_CODE (TREE_OPERAND (expr,0)) == CALL_EXPR)) 
	{
	    tree funcName = TREE_OPERAND (TREE_OPERAND (TREE_OPERAND (expr,0),0),0);
	    
	    /* COME BACK */
	    if (  funcName == lookup_splitc ("__read")
		||funcName == lookup_splitc ("__struct_i_read")
		||funcName == lookup_splitc ("__struct_d_read"))
		return 1;
	    else
		return 0;
	}
	else
	    return 0;

}

/* Strips off the global address from the get calls */
tree
get_global_addr (expr)
     tree expr;
{

    tree funcall;
    tree addr;

    /* Get the function call */
    if ((TREE_CODE (expr) == INDIRECT_REF) &&
	(TREE_CODE (TREE_OPERAND (expr,0)) == CALL_EXPR))
	funcall = TREE_OPERAND (expr, 0);
    else
	funcall = expr;

    addr = TREE_VALUE (TREE_OPERAND (funcall,1));
    return addr;

}

/* Handle ptr + 2 etc .. - should effectively handle all pointer
   calculations - the first input is +/- the second is ptr, third
   is hopefully the number to be added - no checks made on the 3rd input */
tree
build_global_ptr_arith (op, glob_ptr, index)
     enum tree_code op;
     tree glob_ptr, index;
{

    tree size_node;
    tree temp_node;

    glob_ptr = sc_default_conversion (glob_ptr);
    index = sc_default_conversion (index);

    if (TYPE_GLOBAL (TREE_TYPE (glob_ptr))) {
	size_node = parser_build_binary_op (MULT_EXPR,
					    index,
					    c_sizeof (TREE_TYPE (TREE_TYPE (glob_ptr))));

	/* Change for the structure version - AK */
	if (op == MINUS_EXPR)
	    size_node = build_unary_op (NEGATE_EXPR, size_node, 0);
	temp_node = build_address_increments (glob_ptr, size_node);

    } /* else its a split pointer */
    else {
	/* build_split_arith_op just expects an index - so negate index if MINUS_EXPR */
	if (op == MINUS_EXPR)
	    index = build_unary_op (NEGATE_EXPR, index, 0);
	temp_node = build_split_arith_op (glob_ptr, index);
    }
    /* COME BACK */
/*    temp_node = build1 (CONVERT_EXPR, TREE_TYPE (glob_ptr), temp_node); */
    return temp_node;

}

/* Handles += etc for pointers */
tree
build_arith_side_effects (glob_ptr, op, index)
     tree glob_ptr;
     enum tree_code op;
     tree index;
{

    tree size_node;
    tree temp_node;
    tree value;

    glob_ptr = sc_default_conversion (glob_ptr);

    if (op != PLUS_EXPR && op != MINUS_EXPR)
	return error_mark_node;

    if (TYPE_GLOBAL (TREE_TYPE (glob_ptr))) {
	size_node = parser_build_binary_op (MULT_EXPR,
					    index,
					    c_sizeof (TREE_TYPE (TREE_TYPE (glob_ptr))));

	/* Change for the structure version - AK */
	if (op == MINUS_EXPR)
	    size_node = build_unary_op (NEGATE_EXPR, size_node, 0);
	value = build_address_increments (glob_ptr, size_node);
	TREE_TYPE (value) = TREE_TYPE (glob_ptr);

	value = build_modify_expr (glob_ptr, NOP_EXPR, value);
	return value;
    }

    /* Handle x += i; as x = x + i; */
    if (op == MINUS_EXPR)
	index = build_unary_op (NEGATE_EXPR, index, 0);
    
    value = build_split_arith_op (glob_ptr, index);
    TREE_TYPE (value) = TREE_TYPE (glob_ptr);

    value = build_modify_expr (glob_ptr, NOP_EXPR, value);
    return value;
    
}

/* Handles postincrement for global pointers */
tree
build_post_inc (glob_ptr)
     tree glob_ptr;
{

    tree funName;
    tree funCall;
    tree size_node;
    tree ptrNode;
    tree newValue;


    glob_ptr = sc_default_conversion (glob_ptr);

    ptrNode = build_unary_op (ADDR_EXPR, glob_ptr, 0);

    if (TYPE_GLOBAL (TREE_TYPE (glob_ptr))) {
	funName = lookup_splitc ("___post_inc_glob_ptr");
	size_node = c_sizeof (TREE_TYPE (TREE_TYPE (glob_ptr)));
	
	funCall = build_function_call1 (funName, 
					chainon (build_tree_list (NULL_TREE, ptrNode),
						 build_tree_list (NULL_TREE, size_node)));
	
	TREE_TYPE (funCall) = TREE_TYPE (glob_ptr);
	return funCall;
    }

    newValue = build_split_arith_op (glob_ptr, build_int_2 (1, 0));
    funName = lookup_splitc ("___post_assign_spread_ptr");
    funCall = build_function_call1 (funName, 
				    chainon (build_tree_list (NULL_TREE, ptrNode),
					     build_tree_list (NULL_TREE, newValue)));

    TREE_TYPE (funCall) = TREE_TYPE (glob_ptr);
    return funCall;

}

/* Handles postdecrement for global pointers */
tree
build_post_dec (glob_ptr)
     tree glob_ptr;

{

    tree funName = lookup_splitc ("___post_dec_glob_ptr");
    tree funCall;
    tree size_node;
    tree ptrNode;
    tree newValue;

    glob_ptr = sc_default_conversion (glob_ptr);

    ptrNode = build_unary_op (ADDR_EXPR, glob_ptr, 0);
    if (TYPE_GLOBAL (TREE_TYPE (glob_ptr))) {
	size_node = c_sizeof (TREE_TYPE (TREE_TYPE (glob_ptr)));
	
	funCall = build_function_call1 (funName, chainon (build_tree_list (NULL_TREE, ptrNode),
							 build_tree_list (NULL_TREE, size_node)));
	
	TREE_TYPE (funCall) = TREE_TYPE (glob_ptr);
	return funCall;
    }

    newValue = build_split_arith_op (glob_ptr, 
				     build_unary_op (NEGATE_EXPR, build_int_2 (1, 0), 0));
    funName = lookup_splitc ("___post_assign_spread_ptr");
    funCall = build_function_call1 (funName, chainon (build_tree_list (NULL_TREE, ptrNode),
						     build_tree_list (NULL_TREE, newValue)));

    TREE_TYPE (funCall) = TREE_TYPE (glob_ptr);
    return funCall;

}

/* Convert the initial 4byte pointer into a 8 byte value
   but with the same type.
*/
tree
build_initial_split_cast (ptr)
     tree ptr;
{

    tree funName = lookup_splitc ("___spread_cast");
    tree funCall;

    funCall = build_function_call (funName, build_tree_list (NULL_TREE, ptr));
    TREE_TYPE (funCall) = TREE_TYPE (ptr);
    return funCall;
}


/* This is for the index computation - it equals the integer constant 1*/

extern tree size_one_node;

tree
build_split_arith_op (splitPtr, index)
     tree splitPtr, index;
{

    tree incrSize;
    tree unitSize;
    tree element = TREE_TYPE (TREE_TYPE (splitPtr));
    tree temp_ptr;
    tree baseProc, baseAddr;
    tree newProc, newAddr, newIndex;
    tree numChunks;

    splitPtr = sc_default_conversion (splitPtr);
    
    for (temp_ptr = splitPtr; 
	 TREE_CODE (temp_ptr) == CONVERT_EXPR
	 || TREE_CODE (temp_ptr) == NOP_EXPR; )
	temp_ptr = TREE_OPERAND (temp_ptr, 0);

    /* Do folding operation. Get the previous spread pointer
       and its spread index.
    */
    /* Get the proc and address part of the base spread pointer.
       Optimize for the case that the spread pointer was just
       recently built (that is the proc part is 0).
    */
    if (TREE_CODE (temp_ptr) == CALL_EXPR &&
	(TREE_OPERAND (TREE_OPERAND (temp_ptr, 0),0) == 
	 lookup_splitc  ("___create_spread_ptr"))) {

	baseProc = TREE_VALUE (TREE_OPERAND (temp_ptr, 1));
	baseAddr = TREE_VALUE (TREE_CHAIN (TREE_OPERAND (temp_ptr, 1)));
    }
    else if (TREE_CODE (temp_ptr) == CALL_EXPR &&
	     (TREE_OPERAND (TREE_OPERAND (temp_ptr, 0),0) == 
	      lookup_splitc  ("___spread_cast"))) {

	baseProc = build_int_2 (0, 0);
	baseAddr = TREE_VALUE (TREE_OPERAND (temp_ptr, 1));
    }
    else {
	tree funName;

	funName = lookup_splitc  ("toproc");
	baseProc = build_function_call (funName,
					build_tree_list (NULL_TREE, temp_ptr));

	funName = lookup_splitc  ("tolocal");
	baseAddr = build_function_call (funName,
					build_tree_list (NULL_TREE, temp_ptr));
    }

    /* Calculate the size of each unit (in terms of number of
       processors it spans as well as the size in each proc).
    */       
    {
	tree tempIndex, num_elements, type;
	
	type = element;
	num_elements = size_one_node;
	while ((TREE_CODE (type) == ARRAY_TYPE) && TYPE_LANG_FLAG_3 (type)) {
	    tempIndex = TYPE_DOMAIN (type);
	    num_elements = size_binop (MULT_EXPR, num_elements,
				       size_binop (PLUS_EXPR, size_one_node,
						   size_binop (MINUS_EXPR, TYPE_MAX_VALUE (tempIndex),
							       TYPE_MIN_VALUE (tempIndex))));
	    type = TREE_TYPE (type);
	}

	/* The unit size - the block on each proc */
	unitSize = c_sizeof (type);

	/* the number of units per increment */
	incrSize = num_elements;

	/* The new index */
	index = size_binop (MULT_EXPR, index, incrSize);
    }

    /* Index into the base spread  pointer. Generate baseProc + index. */
    newIndex = parser_build_binary_op (PLUS_EXPR, baseProc, index);

    /* Find newProc and newAddr. newProc = newIndex % PROCS (or & PROCS-1).
       newAddr = newIndex / PROCS (or newIndex >> PROCSLOG) * size + baseAddr.
    */
    if (logP)
	newProc = parser_build_binary_op (BIT_AND_EXPR, newIndex,
					  lookup_splitc  ("___PROCSM1"));
    else
	newProc = parser_build_binary_op (TRUNC_MOD_EXPR, newIndex,
					  lookup_splitc  ("___PROCS"));

    if (logP) {
	extern tree integer_type_node;
	TREE_TYPE(newIndex) = integer_type_node;
	numChunks = parser_build_binary_op (RSHIFT_EXPR, newIndex,
					    lookup_splitc  ("___PROCSLOG"));
    }
    else
	numChunks = parser_build_binary_op (TRUNC_DIV_EXPR, newIndex,
					    lookup_splitc  ("___PROCS"));

    /* cast base_addr to a char *. */
    baseAddr = build_c_cast (build_pointer_type (char_type_node), baseAddr);

    /* Add increment to base address. */
    newAddr = parser_build_binary_op (PLUS_EXPR,
				      baseAddr,
				      parser_build_binary_op (MULT_EXPR,
							      numChunks,
							      unitSize));

    /* Build a new split pointer. */
    {
	tree funCall;
	tree funName = lookup_splitc  ("___create_spread_ptr");
	tree paramList = chainon (build_tree_list (NULL_TREE, newProc),
				  build_tree_list (NULL_TREE, newAddr));
	
	funCall = build_function_call (funName, paramList);
	TREE_TYPE (funCall) = TREE_TYPE (splitPtr);
	return funCall;
    }
}



#if 0

tree
build_split_arith_op (splitPtr, index)
     tree splitPtr, index;
{
    tree funName;
    tree funCall;
    tree param_list;
    tree incrSize;
    tree unitSize;
    tree element = TREE_TYPE (TREE_TYPE (splitPtr));
    tree temp_index = index;
    tree prev_index;
    tree temp_ptr;

    splitPtr = sc_default_conversion (splitPtr);

    funName = lookup_splitc ("___index_spread_ptr");

    for (temp_ptr = splitPtr; 
	 TREE_CODE (temp_ptr) == CONVERT_EXPR ||
	 TREE_CODE (temp_ptr) == NOP_EXPR;
	 )
    {
	temp_ptr = TREE_OPERAND (temp_ptr, 0);
    }

    /* Do folding operation */
    if (TREE_CODE (temp_ptr) == CALL_EXPR &&
	(TREE_OPERAND (TREE_OPERAND (temp_ptr, 0),0) == funName)) {
	prev_index = TREE_VALUE (TREE_CHAIN (TREE_OPERAND (temp_ptr, 1)));
	temp_ptr = TREE_VALUE (TREE_OPERAND (temp_ptr, 1));
    }
    else
	prev_index = build_int_2 (0, 0);


    {
	tree index, num_elements, type;
	
	type = element;
	num_elements = size_one_node;
	while ((TREE_CODE (type) == ARRAY_TYPE) && TYPE_LANG_FLAG_3 (type)) {
	    index = TYPE_DOMAIN (type);
	    num_elements = size_binop(MULT_EXPR, num_elements,
				      size_binop(PLUS_EXPR,
						 size_one_node,
						 size_binop(MINUS_EXPR,
						     TYPE_MAX_VALUE (index),
						     TYPE_MIN_VALUE (index))));
	    type = TREE_TYPE (type);
	}

	/* The unit size - the block on each proc */
	unitSize = c_sizeof (type);

	/* the number of units per increment */
	incrSize = num_elements;

    }

    /* split pointer and the index */
    param_list = chainon(build_tree_list(NULL_TREE, temp_ptr),
			 build_tree_list(NULL_TREE,
					 size_binop(PLUS_EXPR,
						    size_binop(MULT_EXPR,
							       temp_index,
							       incrSize),
						    prev_index)));

    param_list = chainon (param_list, build_tree_list (NULL_TREE, unitSize));

    funCall = build_function_call1 (funName, param_list);
    TREE_TYPE (funCall) = TREE_TYPE (splitPtr);

    return funCall;
}
#endif

tree
build_split_ref (splitPtr, index)
     tree splitPtr, index;
{

    tree newPtr;
    tree element = TREE_TYPE (TREE_TYPE (splitPtr));

    newPtr = build_split_arith_op (splitPtr, index);

    if (TREE_CODE (element) != ARRAY_TYPE)
	return build_global_get (newPtr);

    TREE_TYPE (newPtr) = element;
    TREE_LANG_FLAG_3 (newPtr) = 1;
    return newPtr;

}

/* After the access thru the distributed portion of array is made
   we are left with a global pointer - this function deals with
   accessing thru the global pointer the remaining local array part */

tree
build_local_split_ref (globPtr, index)
     tree globPtr, index;
{
    tree element = TREE_TYPE (TREE_TYPE (globPtr));
    tree result;

    TREE_TYPE (globPtr) = c_build_type_variant (build_pointer_type (element), 0, 0, 2);
    result = parser_build_binary_op (PLUS_EXPR, globPtr,
				     parser_build_binary_op (MULT_EXPR, 
							     index, 
							     c_sizeof (element)));

    if (TREE_CODE (element) == ARRAY_TYPE) {
	TREE_TYPE (result) = element;
	TREE_LANG_FLAG_3 (result) = 1;
	return result;
    }

    TREE_TYPE (result) = c_build_type_variant (build_pointer_type (element), 0, 0, 2);
    return build_global_get (result);
					 
}

tree
build_split_ptr_from_array (ptr)
     tree ptr;
{

    tree funName = lookup_splitc ("___spread_cast");
    tree funCall;

    funCall = build_function_call (funName, build_tree_list (NULL_TREE, ptr));
    TREE_TYPE (funCall) = c_build_type_variant (build_pointer_type (TREE_TYPE (TREE_TYPE (ptr))), 
						0, 0, 3);

    return funCall;

}

tree
build_address_increments (glob_ptr, incr)
     tree glob_ptr;
     tree incr;
{

    tree funName = lookup_splitc ("___inc_addr_part");
    tree temp_type = TREE_TYPE (glob_ptr);
    tree funCall;

    funCall = build_function_call1 (funName, chainon (build_tree_list (NULL_TREE, glob_ptr),
						      build_tree_list (NULL_TREE, incr)));

    TREE_TYPE (funCall) = temp_type;
    return funCall;

}

    
tree
build_pointer_compare (op, gPtr1, gPtr2)
     enum tree_code op;
     tree gPtr1, gPtr2;
{

    tree funName;
    tree arg1, arg2;

    if (TREE_CODE (gPtr1) == INTEGER_CST && integer_zerop (gPtr1) ||
	TREE_CODE (gPtr2) == INTEGER_CST && integer_zerop (gPtr2))

	if (op == NE_EXPR)
	    return build_function_call1 (lookup_splitc ("___not_null_ptr_check"),
					 build_tree_list (NULL_TREE,
							  TREE_CODE (gPtr1) == INTEGER_CST ? 
							  gPtr2: gPtr1));
	else
	    return build_function_call1 (lookup_splitc ("___null_ptr_check"),
					 build_tree_list (NULL_TREE,
							  TREE_CODE (gPtr1) == INTEGER_CST ? 
							  gPtr2: gPtr1));

    if (TYPE_GLOBAL_PTR (TREE_TYPE (gPtr1)) && TYPE_GLOBAL_PTR (TREE_TYPE (gPtr2))) {

	arg1 = TREE_TYPE (TREE_TYPE (gPtr1));
	arg2 = TREE_TYPE (TREE_TYPE (gPtr2));
	if (TYPE_MAIN_VARIANT (arg1) == TYPE_MAIN_VARIANT (arg2) ||
	    TREE_CODE (arg1) == VOID_TYPE || TREE_CODE (arg2) == VOID_TYPE) {
	    
	    if (op == NE_EXPR)
		return build_function_call1 (lookup_splitc ("___neq_ptr_check"),
					     chainon (build_tree_list (NULL_TREE, gPtr1),
						      build_tree_list (NULL_TREE, gPtr2)));
	    else
		return build_function_call1 (lookup_splitc ("___eq_ptr_check"),
					     chainon (build_tree_list (NULL_TREE, gPtr1),
						      build_tree_list (NULL_TREE, gPtr2)));

	}

    }
	
    return parser_build_binary_op (op, gPtr1, gPtr2);

}

tree
lookup_splitc_tree(tree name) {

    tree temp;

    if ((temp =lookup_name (name)) != NULL)
	return temp;

    fatal("split-c.h not included.",0);
}	   

/*****************************************/
/* lookup a string defined in split-c.h  */
/*****************************************/

tree
lookup_splitc(const char * const s)
{
    char buffer[256];
    tree id;
	
    id = get_identifier (s);
    id = lookup_name(id);
    if (id) return id;
    sprintf(buffer, "split-c.h not included: '%s' %d.", s, 1);
/*    fatal(buffer, 2); */
    return NULL;
}	   

static int procCount;		/* if ___PROCS used more than once in a
				 * particular split-array definition, 
				 * then complain.
				 */

static tree
foldSplit(tree size)
{
    switch (TREE_CODE(size))
    {
    case MULT_EXPR:
	return size_binop(MULT_EXPR,
			  foldSplit(TREE_OPERAND(size, 0)),
			  foldSplit(TREE_OPERAND(size, 1)));
	break;

    case INTEGER_CST:
	return size;
	break;
	
    case VAR_DECL:
	if (DECL_NAME(size) == get_identifier("___PROCS"))
	{
	    procCount++;
	    return size_one_node;
	}
	else
	    error("Unknown var in split-array decl: %s\n",
		  IDENTIFIER_POINTER(DECL_NAME(size)));
	break;
	
    default:
	error("Array length not an integer multiple of PROCS\n");

#if 0
	printf("foldSplit: finds default:\n");
	debug_tree(size);
#endif
	return size;
    }
}

/* this is only called for static or global level declerations  */

tree
replaceProcsWith1(tree size)
{
    tree temp;
    
#if 0
    printf("replace procs with 1:\n", size, build_numPEs());
    debug_tree(size);
#endif    

    /* first strip off any non_lvalue expr's since we know we don't use it as
     * an lvalue
     */
    
    while (TREE_CODE(size) == NON_LVALUE_EXPR)
	size = TREE_OPERAND(size, 0);

    if (numPes != -1)
    {
	/* make sure it is an integer_cst type */
	if (TREE_CODE (TREE_TYPE (size)) != INTEGER_TYPE
	    && TREE_CODE (TREE_TYPE (size)) != ENUMERAL_TYPE)
	{
	    error ("first dimension of spread-array is non-integer type");
	    return size;
	}

	if (integer_zerop (size))
	    error("First dimension of spread array must be non-zero");

	if ((TREE_CODE (size) != INTEGER_CST))
	{
	    error("First dimension of spread-array must be obvious multiple of PROCS");
	}

	size = size_binop(CEIL_DIV_EXPR, size, build_numPEs());
    }
    else
    {
	/* at the global level remove 1st dimension PROCS */

	procCount = 0;
	size = foldSplit(size);
	if (procCount != 1)
	{
	    error("PROCS used %d times in first dim of spread-array",
		  procCount);
	}
    }
#if 0
    printf("\nreplaced version ->\n");
    debug_tree(size);
#endif    
    return size;
}


void
pushlevel_actions() {

    emit_line_note (input_filename, lineno);
    pushlevel (0);
    clear_last_expr ();
    push_momentary ();
    expand_start_bindings (0);

}

int
setspecs_actions(decl_specs) 
tree decl_specs;
{ 

    int retValue;

    retValue = suspend_momentary ();
    pending_xref_error ();
    declspec_stack = tree_cons (0, current_declspecs,
				declspec_stack);
    current_declspecs = decl_specs;

    return retValue;

}

void
initdcl_actions(declarator)
tree declarator;
{
    tree d = start_decl (declarator, current_declspecs, 0);
    decl_attributes (d, NULL_TREE);
    finish_decl (d, NULL_TREE, NULL_TREE);

}

void
stmt_actions(stmt, file, line)
tree stmt;
char *file;
int  line;
{ 
    stmt_count++;
    emit_line_note (file, line);
    c_expand_expr_stmt (stmt);
    clear_momentary (); 
}

tree
compstmt_actions ()
{
    tree retValue;

    emit_line_note (input_filename, lineno);
    expand_end_bindings (getdecls (), 1, 0);
    retValue = poplevel (1, 1, 0);
    pop_momentary ();
    return retValue;

}

tree
begin_block ()
{

    /* We must force a BLOCK for this level
       so that, if it is not expanded later,
       there is a way to turn off the entire subtree of blocks
       that are contained in it.  */
    keep_next_level ();
    push_label_level ();
    return expand_start_stmt_expr (); 

}

tree
end_block (begin, code)
tree begin;
tree code;
{

    tree rtl_exp;

    pop_label_level ();
    rtl_exp = expand_end_stmt_expr (begin);
    /* The statements have side effects, so the group does.  */
    TREE_SIDE_EFFECTS (rtl_exp) = 1;
    
    /* Make a BIND_EXPR for the BLOCK already made.  */
    return build (BIND_EXPR, TREE_TYPE (rtl_exp),
		  NULL_TREE, rtl_exp, code);

}

/* Transform a write operation of the form "*p = foo();"
   to ({ 
          typeof(foo()) ___temp_var;
	  ___temp_var = foo();
	  __write(p, &___temp_var, sizeof(___temp_var));
	  ___temp_var;
      })
*/
tree
transform_rvalue_writes (lhs, rhs)
tree lhs, rhs;
{

    char *fileName;
    int  lineNo;
    tree expr_type = TREE_TYPE (rhs);
    tree newDeclSpecs;
    int  suspDeclSpecs;
    tree tempVarName;
    tree tempTree;
    tree beginState;


    beginState = begin_block();

    /* simulate pushlevel, save_filename, save_lineno */
    pushlevel_actions();
    fileName = input_filename;
    lineNo = lineno;

    /* typespecs = expr_type
     * reserved_declspecs = NULL_TREE
     * typed_declspecs = tree_cons (NULL_TREE, typespecs, reserved_declspecs
     */
    
    newDeclSpecs = tree_cons (NULL_TREE, expr_type, NULL_TREE);

    /* simulate setspecs actions */
    suspDeclSpecs = setspecs_actions (newDeclSpecs);

    /* ID -> notype_declarator -> declarator */
    tempVarName = get_identifier ("___temp_var");

    /* maybe_asm = NULL_TREE, maybe_attribute = NULL_TREE, form initdcl */
    initdcl_actions (tempVarName);

    /* decl <- typed_declspecs setspecs initdecls */
    current_declspecs = TREE_VALUE (declspec_stack);
    declspec_stack = TREE_CHAIN (declspec_stack);
    resume_momentary (suspDeclSpecs);

    /************* Temp variable declaration is complete **************/

    /* Now build "___temp_var = rhs;" */
    tempTree = build_tree_list (NULL_TREE,
				build_modify_expr (lookup_name (tempVarName),
						   NOP_EXPR, rhs));

    tempTree = build_compound_expr (tempTree);
    stmt_actions (tempTree, fileName, lineNo);

    /* "__write (lhs, &___temp_var, len);" */
    {
	tree funName = lookup_splitc ("__write");
	tree param_list;
	tree source_addr = build_unary_op (ADDR_EXPR,
					   lookup_name (tempVarName),
					   0);

	param_list = chainon (chainon (build_tree_list (NULL_TREE, lhs),
				       build_tree_list (NULL_TREE, source_addr)),
			      build_tree_list (NULL_TREE, 
					       c_sizeof (expr_type)));

	tempTree = build_tree_list (NULL_TREE,
				    build_function_call1 (funName, param_list));
	tempTree = build_compound_expr (tempTree);
	stmt_actions (tempTree, fileName, lineNo);

    }

    /* "___temp_var;" */
    tempTree = build_compound_expr (build_tree_list (NULL_TREE, 
						     lookup_name (tempVarName)));
    stmt_actions (tempTree, fileName, lineNo);

    /* wrap up */
    return end_block (beginState, compstmt_actions());

}

/* Transform reads into something similar. *p is transformed to
   ({
        typeof(*p) ___temp_var;
	__read_local (&___temp_var, p, sizeof(*p));
	___temp_var;
   })
*/
tree
transform_reads (gPtr)
tree gPtr;
{

    char *fileName;
    int  lineNo;
    tree expr_type = TREE_TYPE (TREE_TYPE (gPtr));
    tree newDeclSpecs;
    int  suspDeclSpecs;
    tree tempVarName;
    tree tempTree;
    tree beginState;


    beginState = begin_block();

    /* simulate pushlevel, save_filename, save_lineno */
    pushlevel_actions();
    fileName = input_filename;
    lineNo = lineno;

    /* typespecs = expr_type
     * reserved_declspecs = NULL_TREE
     * typed_declspecs = tree_cons (NULL_TREE, typespecs, reserved_declspecs
     */
    
    newDeclSpecs = tree_cons (NULL_TREE, expr_type, NULL_TREE);

    /* simulate setspecs actions */
    suspDeclSpecs = setspecs_actions (newDeclSpecs);

    /* ID -> notype_declarator -> declarator */
    tempVarName = get_identifier ("___temp_var");

    /* maybe_asm = NULL_TREE, maybe_attribute = NULL_TREE, form initdcl */
    initdcl_actions (tempVarName);

    /* decl <- typed_declspecs setspecs initdecls */
    current_declspecs = TREE_VALUE (declspec_stack);
    declspec_stack = TREE_CHAIN (declspec_stack);
    resume_momentary (suspDeclSpecs);

    /************* Temp variable declaration is complete **************/


    /* "__read_local (&___temp_var, p, len);" */
    {
	tree funName = lookup_splitc ("__read_local");
	tree param_list;
	tree source_addr = build_unary_op (ADDR_EXPR,
					   lookup_name (tempVarName),
					   0);

	param_list = chainon (chainon (build_tree_list (NULL_TREE, source_addr),
				       build_tree_list (NULL_TREE, gPtr)),
			      build_tree_list (NULL_TREE, 
					       c_sizeof (expr_type)));

	tempTree = build_tree_list (NULL_TREE,
				    build_function_call1 (funName, param_list));
	tempTree = build_compound_expr (tempTree);
	stmt_actions (tempTree, fileName, lineNo);

    }

    /* "___temp_var;" */
    tempTree = build_compound_expr (build_tree_list (NULL_TREE, 
						     lookup_name (tempVarName)));
    stmt_actions (tempTree, fileName, lineNo);

    /* wrap up */
    return end_block (beginState, compstmt_actions());

}

/* Check whether there is a structure read operation */
int CheckReadOp(expr)
tree expr;
{

    if ((TREE_CODE (expr) == INDIRECT_REF) &&
	(TREE_CODE (TREE_OPERAND (expr,0)) == CALL_EXPR)) {
	
	tree funcName = TREE_OPERAND (TREE_OPERAND (TREE_OPERAND (expr,0),0),0);
	if (  funcName == lookup_splitc ("__read") )
	    return 1;

    }
    
    return 0;  

}

/* Extract the global pointer assuming its a read op */
tree ExtractGlobalPtr(expr)
tree expr;
{

    return TREE_VALUE (TREE_OPERAND (TREE_OPERAND (expr, 0), 1));

}

/* Recurse through an expression chain and convert all
   __read calls to the complex C stmt that mimics it.  */
tree
modify_expr_list(exprList)
tree exprList;
{

    tree tempExpr;
    tree next;

    if (exprList == NULL_TREE) return exprList;

    tempExpr = TREE_VALUE (exprList);
    next = TREE_CHAIN (exprList);

    if (CheckReadOp (tempExpr))
	TREE_VALUE (exprList) = transform_reads (ExtractGlobalPtr (tempExpr));
    else
	TREE_VALUE (exprList) = sc_default_conversion (tempExpr);

    TREE_CHAIN (exprList) = modify_expr_list (next);

    return exprList;
	
}

/* Transform *p ++; or *p --; to something equivalent. */
tree
transform_side_effect_stmt (gPtr, op)
tree gPtr;
enum tree_code op;
{

    char *fileName;
    int  lineNo;
    tree expr_type = TREE_TYPE (TREE_TYPE (gPtr));
    tree newDeclSpecs;
    int  suspDeclSpecs;
    tree tempVarName;
    tree tempTree;
    tree beginState;

    beginState = begin_block();

    /* simulate pushlevel, save_filename, save_lineno */
    pushlevel_actions();
    fileName = input_filename;
    lineNo = lineno;

    /* typespecs = expr_type
     * reserved_declspecs = NULL_TREE
     * typed_declspecs = tree_cons (NULL_TREE, typespecs, reserved_declspecs
     */
    
    newDeclSpecs = tree_cons (NULL_TREE, expr_type, NULL_TREE);

    /* simulate setspecs actions */
    suspDeclSpecs = setspecs_actions (newDeclSpecs);

    /* ID -> notype_declarator -> declarator */
    tempVarName = get_identifier ("___temp_var");

    /* maybe_asm = NULL_TREE, maybe_attribute = NULL_TREE, form initdcl */
    initdcl_actions (tempVarName);

    /* decl <- typed_declspecs setspecs initdecls */
    current_declspecs = TREE_VALUE (declspec_stack);
    declspec_stack = TREE_CHAIN (declspec_stack);
    resume_momentary (suspDeclSpecs);

    /************* Temp variable declaration is complete **************/

    /* "___temp_var = *p;" */
    tempTree = build_tree_list (NULL_TREE,
				build_modify_expr (lookup_name (tempVarName),
						   NOP_EXPR,
						   build_global_get (gPtr)));
    tempTree = build_compound_expr (tempTree);
    stmt_actions (tempTree, fileName, lineNo);

    /* "*p = ___temp_var op 1;" */
    tempTree = parser_build_binary_op (op, lookup_name (tempVarName), 
				       size_one_node);
    tempTree = build_tree_list (NULL_TREE,
				build_global_put (gPtr, tempTree, 0));
    tempTree = build_compound_expr (tempTree);
    stmt_actions (tempTree, fileName, lineNo);

    /* "___temp_var;" */
    tempTree = build_compound_expr (build_tree_list (NULL_TREE, 
						     lookup_name (tempVarName)));
    stmt_actions (tempTree, fileName, lineNo);

    /* wrap up */
    return end_block (beginState, compstmt_actions());

}

/* Default conversions. From spread array to
   a spread pointer. From a array whose tree_lang_flag_3
   is set to a global pointer?
*/
tree
sc_default_conversion (expr)
    tree expr;
{
    tree element;

    /* If a 8 byte spread array, make it a 8-byte spread pointer. */
    if (TYPE_SPREAD_ARRAY (TREE_TYPE (expr))) {
	element = TREE_TYPE (TREE_TYPE (expr));
	TREE_TYPE (expr) = c_build_type_variant (build_pointer_type(element),
						 0, 0, 3);
	return expr;
    }

    /* If an array formed by global pointer dereferences,
       cast into a global pointer.
    */
    if (TREE_CODE (TREE_TYPE (expr)) == ARRAY_TYPE
	&& TREE_LANG_FLAG_3 (expr)) {
	element = TREE_TYPE (TREE_TYPE (expr));
	TREE_TYPE (expr) = c_build_type_variant (build_pointer_type(element),
						 0, 0, 2);
	return expr;
    }
    return expr;

}

/* Insert checks for "+", and "-" ops on global
   pointers. Generate some errors also.
*/
tree
sc_parser_build_binary_op (code, op1, op2)
     enum tree_code code;
     tree op1, op2;
{

    op1 = sc_default_conversion(op1);
    op2 = sc_default_conversion(op2);

    /* First check for the operations that needs to be transformed. */
    if (TYPE_GLOBAL_PTR (TREE_TYPE (op1))
	|| TYPE_GLOBAL_PTR (TREE_TYPE (op2))) {

	if (code == PLUS_EXPR || code == MINUS_EXPR) {

	    /* Check for addition/subtraction of two pointers. */
	    if (TYPE_GLOBAL_PTR (TREE_TYPE (op1))
		&& TYPE_GLOBAL_PTR (TREE_TYPE (op2))) {
		error("Invalid operands to binary operator.");
		return error_mark_node;
	    }

	    if (TYPE_GLOBAL_PTR (TREE_TYPE (op1)))
		return build_global_ptr_arith (code, op1, op2);
	    else
		return build_global_ptr_arith (code, op2, op1);

	}

	if (code == EQ_EXPR	|| code == NE_EXPR)
	    return build_pointer_compare (code, op1, op2);

	error("Invalid operands to binary operator.");
    }
    return parser_build_binary_op(code, op1, op2);

}

/* The entry point for building unary functions. */
tree
sc_build_unary_op (op, expr)
     enum tree_code op;
     tree expr;
{

    expr = sc_default_conversion (expr);
    if ((op != ADDR_EXPR) || !global_op (expr)) {
	/* Handle ++p and --q on global/spread pointers. */

	if ((op == PREINCREMENT_EXPR || op == PREDECREMENT_EXPR)
	    && TYPE_GLOBAL_PTR (TREE_TYPE (expr))) {

	    if (op == PREINCREMENT_EXPR)
		return build_modify_expr (expr, NOP_EXPR,
					  build_global_ptr_arith (PLUS_EXPR,
								  expr,
								  build_int_2 (1, 0)));
	    else
		return build_modify_expr (expr, NOP_EXPR,
					  build_global_ptr_arith (MINUS_EXPR,
								  expr,
								  build_int_2 (1, 0)));
	}
	else if (global_op (expr) 
		 && (op == PREINCREMENT_EXPR || op == PREDECREMENT_EXPR)) {
	    /* Handle ++(*p) or --(*p). */

	    tree gPtr = get_global_addr (expr);
	    tree gVal;

	    gVal = build_global_get (gPtr);
	    gVal = parser_build_binary_op ((op == PREINCREMENT_EXPR
					    ? PLUS_EXPR : MINUS_EXPR),
					   gVal, build_int_2 (1, 0));
	    return build_global_put (gPtr, gVal, 0);
	}
	else {
	    /* The normal C case. */
	    tree temp = build_unary_op (op, expr, 0); 
	    overflow_warning (temp);
	    return temp;
	}
    }
    else
	return get_global_addr (expr);
	
}

/* Handle the regular "=" case with possible rhs/lhs global references. */
tree
sc_build_blocking_ops (lhs, op, rhs)
     tree lhs;
     enum tree_code op;
     tree rhs;
{ 
    /* Check whether lhs or rhs is a indirect-ref and then translate them into
       gets or puts appropriately - actually should do it for -> also */
    rhs = sc_default_conversion(rhs);
    
    if (TREE_CODE (lhs) == CALL_EXPR) {
	if (TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc ("__c_read")
	    ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc  ("__sh_read")
	    ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc ("__i_read")
	    ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc  ("__ll_read")
	    ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc ("__f_read")
	    ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc ("__d_read"))
	    return build_global_put (TREE_VALUE (TREE_OPERAND (lhs, 1)), rhs, 0);
	else
	    return build_modify_expr_with_opt (lhs, NOP_EXPR, rhs, 0);
    }
    else
	if ((TREE_CODE (lhs) == INDIRECT_REF) &&
	    (TREE_CODE (TREE_OPERAND (lhs,0)) == CALL_EXPR))
	    {
		tree funcName = TREE_OPERAND (TREE_OPERAND (TREE_OPERAND (lhs,0),0),0);
		
		if (funcName==lookup_splitc("__read") ||
		    funcName==lookup_splitc("__struct_i_read")||
		    funcName==lookup_splitc("__struct_d_read"))
		    return build_global_put(TREE_VALUE(TREE_OPERAND (TREE_OPERAND (lhs,0), 1)),
					  rhs, 0);
		else
		    return build_modify_expr_with_opt (lhs, NOP_EXPR, rhs, 0);
		
	    }
	else
	    return build_modify_expr_with_opt (lhs, NOP_EXPR, rhs, 0); 
    
}



tree
sc_build_nonblocking_ops(lhs, op, rhs)
     tree lhs;
     enum tree_code op;
     tree rhs;
{
    /* Copied the piece of code from prev segment */
    /* Check whether lhs or rhs is a indirect-ref and then translate them into
	gets or puts appropriately - actually should do it for -> also */
    rhs = sc_default_conversion(rhs);
    if (TREE_CODE (lhs) == CALL_EXPR) {
	if (  TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc ("__c_read")
	    ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc ("__sh_read")
	    ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc ("__i_read")
	    ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc ("__ll_read")
	    ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc ("__f_read")
	    ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	    lookup_splitc ("__d_read"))
	    return build_global_put (TREE_VALUE (TREE_OPERAND (lhs, 1)), rhs, 1);
	else
	    return build_modify_expr_with_opt (lhs, NOP_EXPR, rhs, 1);
    }
    else
	if ((TREE_CODE (lhs) == INDIRECT_REF) &&
	    (TREE_CODE (TREE_OPERAND (lhs,0)) == CALL_EXPR))
	    {
		tree funcName = TREE_OPERAND (TREE_OPERAND (TREE_OPERAND (lhs,0),0),0);
		
		if (  funcName == lookup_splitc ("__read") ||
		    funcName == lookup_splitc ("__struct_i_read") ||
		    funcName == lookup_splitc ("__struct_d_read"))
		    return build_global_put (TREE_VALUE (TREE_OPERAND (TREE_OPERAND (lhs,0), 1)),
					   rhs, 1);
		else
		    return build_modify_expr_with_opt (lhs, NOP_EXPR, rhs, 1);
		
	    }
	else
	    return build_modify_expr_with_opt (lhs, NOP_EXPR, rhs, 1); 
}

tree
sc_build_store_ops (lhs, op, rhs)
     tree lhs;
     enum tree_code op;
     tree rhs;
{
    tree retVal;

    rhs = sc_default_conversion(rhs);
    if ((TREE_CODE (lhs) == CALL_EXPR) &&
	(  TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	 lookup_splitc ("__c_read")
	 ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	 lookup_splitc ("__sh_read")
	 ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	 lookup_splitc ("__i_read")
	 ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	 lookup_splitc ("__ll_read")
	 ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	 lookup_splitc ("__f_read")
	 ||TREE_OPERAND (TREE_OPERAND (lhs, 0),0) == 
	 lookup_splitc ("__d_read")))
	retVal = build_global_put (TREE_VALUE (TREE_OPERAND (lhs, 1)), rhs, 2);
    else if (TREE_CODE (lhs) == INDIRECT_REF &&
	     TREE_CODE (TREE_OPERAND (lhs,0)) == CALL_EXPR &&
	     (  TREE_OPERAND (TREE_OPERAND (TREE_OPERAND (lhs, 0), 0), 0) == 
	      lookup_splitc ("__read")
	      ||TREE_OPERAND (TREE_OPERAND (TREE_OPERAND (lhs, 0), 0), 0) == 
	      lookup_splitc ("__struct_i_read")
	      ||TREE_OPERAND (TREE_OPERAND (TREE_OPERAND (lhs, 0), 0), 0) == 
	      lookup_splitc ("__struct_d_read")))
	retVal = build_global_put (TREE_VALUE (TREE_OPERAND (TREE_OPERAND (lhs, 0), 1)), rhs, 2);
    else /* Actually this can be directly branded as an error */
	retVal = build_modify_expr (lhs, NOP_EXPR, rhs);
    C_SET_EXP_ORIGINAL_CODE (retVal, MODIFY_EXPR);
    return retVal;

}

tree
sc_build_assign_ops (lhs, op, rhs)
     tree lhs;
     enum tree_code op;
     tree rhs;
{
    tree retVal;

    if (TREE_TYPE (TREE_TYPE (lhs)) && TYPE_GLOBAL_PTR (TREE_TYPE (lhs)))
	retVal = build_arith_side_effects (lhs, op, rhs);
    else if (global_op (lhs)) {
	/* rewrite it as *p = *p OP expr; */
	tree gPtr = get_global_addr (lhs);
	tree temp;
	
	temp = build_global_get (gPtr);
	
	temp = parser_build_binary_op (op, temp, rhs);
	
	/* An hack - absolutely disgusting one - no more*/
	retVal = build_global_put (gPtr, temp, 0);
#ifdef NOMORE
	retVal = error_mark_node;
	error ("Not currently supported - rewrite the statement");
#endif
    }
    else
	retVal = build_modify_expr (lhs, op, rhs); 
    /* This inhibits warnings in truthvalue_conversion.  */
    C_SET_EXP_ORIGINAL_CODE (retVal, ERROR_MARK);

    return retVal;

}

tree
sc_sizeof (type)
     tree type;
{
    if (TYPE_SPREAD_ARRAY (type))
	return size_binop(MULT_EXPR,
			  lookup_splitc("___PROCS"),
			  c_sizeof(type));
    else
	return c_sizeof(type);
    
}
